#!/usr/bin/env python


from math import ceil

import random
import sys

_colours = ["black", "white", "red", "green", "blue", "mauve"]
_woods = ["oak", "pine", "cherry", "teak", "mahogany", "beech", "walnut"]
_machines = [
    "highspeed-saw",
    "glazer",
    "grinder",
    "immersion-varnisher",
    "planer",
    "saw",
    "spray-varnisher",
]
_boardsize_prefix = "s"


def default_machines(num_machines):
    return {machine: num_machines for machine in _machines}


class Part:
    def __init__(self, index, woods, colours, problemtype=None):
        self.name = "p%d" % index
        self.problemtype = problemtype
        self.generate_random_goal(woods, colours)
        self.generate_random_init(woods, colours)

    def generate_random_goal(self, woods, colours):
        self.goalprops = dict()
        self.goalprops["treatment"] = random.choice(["varnished", "glazed"])
        self.goalprops["surface-condition"] = random.choice(["smooth", "verysmooth"])

        if self.problemtype != "painting":
            colours = colours + ["natural"]

        self.goalprops["colour"] = random.choice(colours)
        self.goalprops["wood"] = random.choice(woods)
        self.generate_goal_selection()

    def generate_goal_selection(self, nr_goals=None):
        nr_goals = random.choice([2, 2, 2, 3, 4])
        self.goalselection = set(random.sample(list(self.goalprops), nr_goals))

    def generate_random_init(self, woods, colours):
        def gen_preprocessing_status():
            self.initprops["treatment"] = random.choice(
                ["varnished", "glazed", "colourfragments"]
            )
            self.initprops["surface-condition"] = random.choice(
                ["rough", "smooth", "verysmooth"]
            )
            self.initprops["wood"] = self.goalprops["wood"]

            poss_colours = set(colours + ["natural"]) - {self.goalprops["colour"]}
            self.initprops["colour"] = random.choice(list(poss_colours))

        self.initprops = dict()

        status = random.choice(["unused"] * 4 + ["available"])
        if status != "unused":
            while True:
                gen_preprocessing_status()
                goal_initially_satisfied = True
                for prop in self.goalselection:
                    if self.initprops[prop] != self.goalprops[prop]:
                        goal_initially_satisfied = False
                        break

                if not goal_initially_satisfied:
                    break

                self.generate_goal_selection(len(self.goalselection))

        self.size = random.randint(1, 3)

    def dump_init(self, indent="", out=None):
        if self.initprops:
            print(f"{indent}(available {self.name})", file=out)

            for prop, val in self.initprops.items():
                print(f"{indent}({prop} {self.name} {val})", file=out)
        else:
            print(f"{indent}(unused {self.name})", file=out)

        print(
            "%s(goalsize %s %s)"
            % (indent, self.name, ["small", "medium", "large"][self.size - 1]),
            file=out,
        )
        print(
            f"{indent}(= (spray-varnish-cost {self.name}) {self.size * 5})",
            file=out,
        )
        print(
            "%s(= (glaze-cost %s) %d)" % (indent, self.name, self.size * 5 + 5),
            file=out,
        )
        print(
            "%s(= (grind-cost %s) %d)" % (indent, self.name, self.size * 15), file=out
        )
        print(
            "%s(= (plane-cost %s) %d)" % (indent, self.name, self.size * 10), file=out
        )

    def dump_goal(self, indent="", out=None):
        print(f"{indent}(available {self.name})", file=out)

        for choice in self.goalselection:
            print(
                f"{indent}({choice} {self.name} {self.goalprops[choice]})",
                file=out,
            )


class Board:
    def __init__(self, index, wood, size):
        self.wood = wood
        self.name = "b%d" % index
        self.size = size
        self.surface = random.choice(["rough", "rough", "rough", "smooth"])

    def dump_init(self, indent="", out=None):
        print(
            "%s(boardsize %s %s%d)" % (indent, self.name, _boardsize_prefix, self.size),
            file=out,
        )
        print(f"{indent}(wood {self.name} {self.wood})", file=out)
        print(
            f"{indent}(surface-condition {self.name} {self.surface})", file=out
        )
        print(f"{indent}(available {self.name})", file=out)


class Machine:
    def __init__(self, name, type):
        self.name = name
        self.type = type
        self.colours = set()

    def dump_init(self, indent, out=None):
        if self.type == "highspeed-saw":
            print(f"{indent}(empty {self.name})", file=out)

        for colour in self.colours:
            print(f"{indent}(has-colour {self.name} {colour})", file=out)


class Task:
    def __init__(
        self,
        nr_parts,
        nr_machines,
        wood_factor=1.0,
        problemtype=None,
        **additional_machines
    ):
        self.problemtype = problemtype
        self.wood_factor = wood_factor

        nr_colours = max(int(round(0.7 * nr_parts)), 2)
        self.colours = random.sample(_colours, min(nr_colours, len(_colours)))

        nr_woods = max(int(round(0.25 * nr_parts)), 2)
        self.woods = random.sample(_woods, min(nr_woods, len(_woods)))

        self.parts = [
            Part(nr, self.woods, self.colours, problemtype) for nr in range(nr_parts)
        ]

        self.max_board_size = self._generate_boards()

        self._generate_machines(nr_machines, **additional_machines)
        self._assign_colours_to_machines()

    def _generate_boards(self):
        quantities = dict()

        for part in [p for p in self.parts if not p.initprops]:
            wood = part.goalprops["wood"]
            quantities.setdefault(wood, []).append(part.size)

        self.boards = []
        counter = 0
        maxsize = 0
        for wood, quantity in quantities.items():
            if not quantity:
                self.woods.remove("wood")
            else:
                additional_wood = int(ceil(sum(quantity) * (self.wood_factor - 1)))

                while additional_wood > 3:
                    rand = random.randint(1, 3)
                    quantity.append(rand)
                    additional_wood -= rand

                quantity.append(additional_wood)
                random.shuffle(quantity)
                for amounts in zip(
                    quantity[::4],
                    quantity[1::4] + [0],
                    quantity[2::4] + [0],
                    quantity[3::4] + [0],
                ):
                    size = sum(amounts)
                    maxsize = max(maxsize, size)
                    self.boards.append(Board(counter, wood, size))
                    counter += 1

        return maxsize

    def _generate_machines(self, nr_machines, **changes):
        machines = dict(default_machines(nr_machines))
        machines.update(changes)

        self.machines = dict()
        for type, number in machines.items():
            if number:
                m = [Machine(f"{type}{nr}", type) for nr in range(number)]
                self.machines[type] = m

    def _assign_colours_to_machines(self):
        necessary = self._determine_necessary_colours()

        spray_varnishers = self.machines.get("spray-varnisher", [])
        immersion_varnishers = self.machines.get("immersion-varnisher", [])
        glazers = self.machines.get("glazer", [])

        for colour in necessary["varnished"]:
            machines = random.sample(
                spray_varnishers, int(ceil(0.6 * len(spray_varnishers)))
            )
            machines += random.sample(
                immersion_varnishers, int(ceil(0.6 * len(immersion_varnishers)))
            )

            for m in machines:
                m.colours.add(colour)

        for colour in necessary["glazed"]:
            for m in random.sample(glazers, int(ceil(0.6 * len(glazers)))):
                m.colours.add(colour)

        for colour in necessary["unspecified"]:
            machines = random.sample(
                spray_varnishers, int(ceil(0.3 * len(spray_varnishers)))
            )
            machines += random.sample(
                immersion_varnishers, int(ceil(0.3 * len(immersion_varnishers)))
            )
            machines += random.sample(glazers, int(ceil(0.3 * len(glazers))))

            for m in machines:
                m.colours.add(colour)

        for type in ["spray-varnisher", "immersion-varnisher", "glazer"]:
            for m in self.machines.get(type, []):
                if not m.colours:
                    m.colours.add(random.choice(list(self.colours)))

    def _determine_necessary_colours(self):
        used_colours = dict()
        used_colours["varnished"] = set()
        used_colours["glazed"] = set()
        used_colours["unspecified"] = set()

        for part in self.parts:
            if "colour" in part.goalselection:
                colour = part.goalprops["colour"]

                if "treatment" in part.goalselection:
                    used_colours[part.goalprops["treatment"]].add(colour)
                else:
                    used_colours["unspecified"].add(colour)

        return used_colours

    def dump(self, out=None):
        self._dump_header(out)
        self._dump_objects("  ", out)
        self._dump_init("  ", out)
        self._dump_goal("  ", out)
        self._dump_metric("  ", out)

        print(")", file=out)

    def _dump_header(self, out=None):
        print(
            "; woodworking task with %s parts and %d%% wood"
            % (len(self.parts), self.wood_factor * 100),
            file=out,
        )
        print("; Machines:", file=out)

        for type, machines in self.machines.items():
            print(";   %d %s" % (len(machines), type), file=out)

        print("", file=out)
        print("(define (problem wood-prob)", file=out)
        print("  (:domain woodworking)", file=out)

    def _dump_objects(self, indent="", out=None):
        print(indent + "(:objects", file=out)

        for type, machines in self.machines.items():
            print(
                "{}  {} - {}".format(indent, " ".join([m.name for m in machines]), type),
                file=out,
            )

        print("{}  {} - acolour".format(indent, " ".join(self.colours)), file=out)
        print("{}  {} - awood".format(indent, " ".join(self.woods)), file=out)
        print(
            "{}  {} - part".format(indent, " ".join([p.name for p in self.parts])), file=out
        )
        print(
            "{}  {} - board".format(indent, " ".join([b.name for b in self.boards])),
            file=out,
        )

        boardsizes = [
            "%s%d" % (_boardsize_prefix, size)
            for size in range(self.max_board_size + 1)
        ]

        print("{}  {} - aboardsize".format(indent, " ".join(boardsizes)), file=out)
        print(indent + ")", file=out)

    def _dump_init(self, indent="", out=None):
        print(indent + "(:init", file=out)
        print(indent + "  (grind-treatment-change varnished colourfragments)", file=out)
        print(indent + "  (grind-treatment-change glazed untreated)", file=out)
        print(indent + "  (grind-treatment-change untreated untreated)", file=out)
        print(indent + "  (grind-treatment-change colourfragments untreated)", file=out)
        print(indent + "  (is-smooth smooth)", file=out)
        print(indent + "  (is-smooth verysmooth)", file=out)
        print(indent + "  (= (total-cost) 0)", file=out)

        for size in range(self.max_board_size):
            print(
                "%s  (boardsize-successor %s%d %s%d)"
                % (indent, _boardsize_prefix, size, _boardsize_prefix, size + 1),
                file=out,
            )

        for type in self.machines:
            for m in self.machines[type]:
                m.dump_init(indent + "  ", out)

        for part in self.parts:
            part.dump_init(indent + "  ", out)

        for board in self.boards:
            board.dump_init(indent + "  ", out)

        print(indent + ")", file=out)

    def _dump_goal(self, indent="", out=None):
        print(indent + "(:goal", file=out)
        print(indent + "  (and", file=out)

        for part in self.parts:
            part.dump_goal(indent + "  ", out)

        print(indent + "  )", file=out)
        print(indent + ")", file=out)

    def _dump_metric(self, indent="", out=None):
        print(indent + "(:metric minimize (total-cost))", file=out)


if __name__ == "__main__":
    if len(sys.argv) not in [4, 5]:
        print(
            "Usage: python create_woodworking_instance.py "
            "wood_factor size num_machines [random_seed]"
        )
        sys.exit(1)

    wood_factor = float(sys.argv[1])
    size = int(sys.argv[2])
    num_machines = int(sys.argv[3])

    if len(sys.argv) == 5:
        seed = int(sys.argv[4])
    else:
        seed = None

    random.seed(seed)

    instance = Task(size, num_machines, wood_factor)
    instance.dump(out=sys.stdout)



#! /usr/bin/env python

from rolling_stone_data import rs_data

rs_nodes = {
    instance: nodes for instance, moves, pushes, nodes in rs_data if moves is not None
}


def r(begin, end=None):
    if end is None:
        end = begin
    return set(range(begin, end + 1))


solved_by_base = (
    r(1, 34)
    | r(36, 77)
    | r(79, 92)
    | r(94, 97)
    | r(100, 104)
    | r(110)
    | r(115, 116)
    | r(119, 120)
    | r(124, 125)
    | r(127, 132)
    | r(134, 136)
    | r(142)
    | r(147, 149)
    | r(151, 152)
    | r(154, 155)
)


def key(instance):
    hard = instance not in solved_by_base
    too_hard = instance not in rs_nodes
    return (hard, too_hard, rs_nodes.get(instance))


instances = list(range(1, 156))
instances.sort(key=key)

for no, instance in enumerate(instances):
    print("%3d  instance #%03d: %s" % ((no + 1), instance, key(instance)))

assert len(instances) == 155, len(instances)
assert len(solved_by_base) == 123, len(solved_by_base)
assert len(rs_nodes) == 145, len(rs_nodes)
assert set(solved_by_base - set(rs_nodes)) == {154}

# "instances" now contains, in sequence:
# 122 instances solved by blind search and Rolling Stone.
#   1 instance (#154) solved by blind search, but not by Rolling Stone.
#  23 instances solved by Rolling Stone, but not blind search.
#   9 instance not solved by anyone.
#
# Within the first and third range, instances are sorted by ascending
# node expansion count for Rolling Stone (ties broken by original
# problem number). Within the fourth range, they are ordered by
# original problem number.

opt_easy = instances[3:113:11]
opt_middle = instances[113:123]
opt_hard = instances[123:133]

opt = opt_easy + opt_middle + opt_hard

sat_easy = opt_easy[1::2]
sat_middle = opt_middle[1::2]
sat_hard = instances[123:143]

sat = sat_easy + sat_middle + sat_hard

print("sat_range = [%s]" % ", ".join(map(str, sat)))
print("opt_range = [%s]" % ", ".join(map(str, opt)))



""" File:        cmd_line.py
    Author:      Nathan Robinson
    Contact:     nathan.m.robinson@gmail.com
    Date:        2012-07-07
    Desctiption: This module describes a generic command line input processor
                 for python programs

    Liscence:   Copyright (c) Year 2012, Nathan Robinson <nathan.m.robinson@gmail.com>

                Permission to use, copy, modify, and/or distribute this software for any
                purpose with or without fee is hereby granted, provided that the above
                copyright notice and this permission notice appear in all copies.

                THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
                WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
                MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
                SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
                WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
                ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
                IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
"""

import sys

# Preliminary definitions
# -------------------------------------------------------------------------------


class InputException(Exception):
    """A custom Exception thrown by ArgProcessor.parse_args(), et al."""


class ArgDefinition:
    """This class stores the definition of a command line argument that can be
    passed to the program.
    """

    def __init__(
        self, var_name, needed, validator, validator_args, default_value, description
    ):
        """Create a new ArgDefinition.

        (ArgDefinition, str, bool, function, [object], object, str) -> None
        var_name: The name of the variable where the value will be stored.
        needed: If the argument can be is required (default not used then).
        validator: A function pointer that will take the read in and
            validate the value of this argument and either throw an
            InputException or return the parsed value (None for no
            validation).
        validator_args: A list of the arguments (except the value) for the
            validator function.
        default_value: The value that the var_name will be set to if one is
            not given on the commane line. This will not be validated.
        description: A string describing the argument to show after parsing
            if the the verbose option is selected (it is by default).
        """

        self.var_name = var_name
        self.needed = needed
        self.validator = validator
        self.validator_args = validator_args
        self.default_value = default_value
        self.description = description


class FlagDefinition:
    """This class stores the definition of a command line flag that can be
    passed to the program.
    """

    def __init__(self, var_name, function, description):
        """Create a new FlagDefinition.

        (FlagDefinition, str, function, str) -> None
        var_name: The name of the variable where the value will be stored.
        function: A pointer to a function that will be called if the flag is
            present (or None). Flag functions are called in the order that
            flags are defined in, before arguments are passed.
            The function must take a single argument, the ArgProcessor.
        description: A string describing the flag to show after parsing
            unless the suppress flag is supplied.
        """

        self.var_name = var_name
        self.function = function
        self.description = description


# Validators
# A set of functions that validate the various types of command line arguments.
# -------------------------------------------------------------------------------


def range_validator(value_str, args):
    """Validate command line args that take a range of values.

    (str, [type, object, object, bool, str]) -> object

    value_str: The command line string representing the argument.
    args: A list containing the parameters that the validator requires:
          0 - type:       The python type of the cmd line arg;
          1 - lb:         The lower bound of the cmd line arg (or None);
          2 - ub:         The upper bound of the cmd line arg (or None);
          3 - allow_none: If the arg can have the value 'None'; and
          4 - error_msg:  The error message to be printed (followed by the
                          value) if the supplied value is invalid.
    """

    assert len(args) == 5, "Error: range_validator requires 5 arguments."
    a_type, lb, ub, allow_none, error_msg = args
    try:
        if allow_none and value_str == "None":
            value = None
        else:
            value = a_type(value_str)
    except ValueError:
        raise InputException(error_msg + " " + value_str)
    if (lb is not None and value < lb) or (ub is not None and value > ub):
        raise InputException(error_msg + " " + value_str)
    return value


def range_validator_advice(validator_args):
    """A function for the range_validator that returns a short help string
    based on the supplied args.

    ([type, object, object, bool, str]) -> str
    """

    a_type, lb, ub, allow_none, error_msg = validator_args
    if lb is None and ub is None:
        return ""
    adv_str = "x"
    if lb is not None:
        adv_str = str(lb) + " <= " + adv_str
    if ub is not None:
        adv_str += " <= " + str(ub)
    if allow_none:
        adv_str += ", None"
    return " {" + adv_str + "}"


def enum_validator(value_str, args):
    """Validate cmd line args that take one of a set of str values.

    (str, [[str], str]) -> str
    value_str: The command line string representing the argument.
    args:      A list containing the parameters that the validator requires:
        0 - v_list: The list of valid string values that the cmd line arg
                    can take; and
        1 - error_msg: The error message to be printed (followed by the
                       value) if the supplied value is invalid.
    """

    assert len(args) == 2, "Error: enum_validator requires 2 arguments."
    v_list, error_message = args
    if v_list is not None and value_str not in v_list:
        raise InputException(error_message + " " + value_str)
    return value_str


def enum_validator_advice(validator_args):
    """A function for the enum_validator that returns a short help string based
    on the supplied args.

    ([[str], str]) -> str
    """

    return " {" + ", ".join(validator_args[0]) + "}"


def bool_validator(value_str, args):
    """Validate Boolean command line args.

    (str, [str]) -> bool
    value_str: The command line string representing the arg.
    args:      A list containing the parameters that the validator requires:
               0 - error_msg: The error message to be printed (followed by
                              the value) if the supplied value is invalid.
    """

    assert len(args) == 1, "Error: bool_validator requires 1 argument."
    error_message = args[0]
    if value_str == "True":
        value = True
    elif value_str == "False":
        value = False
    else:
        raise InputException(error_message + " " + value_str)
    return value


def bool_validator_advice(validator_args):
    """A function for the bool_validator that returns a short help string

    ([str]) -> str
    """

    return " {True, False}"


def seq_range_validator(value_str, args):
    """Validate a sequence of command line args that take a range of values,
    each value is seperated by the specified seperator.

    (str, [str, type, object, object, bool, str]) -> [Number]

    value_str: The command line string representing the argument.
    args: A list containing the parameters that the validator requires:
          0 - sep:        The seperator used to separate the list of values
          1 - min_vals:   The minimum number of values (or None);
          2 - max_vals:   The maximum number of values (or None);
          3 - type:       The python type of the cmd line arg;
          4 - lb:         The lower bound of the values (or None);
          5 - ub:         The upper bound of the values (or None);
          6 - error_msg:  The error message to be printed (followed by the
                          value) if the supplied value is invalid.
    """

    assert len(args) == 7, "Error: seq_range_validator requires 7 arguments."
    sep, min_vals, max_vals, a_type, lb, ub, error_msg = args
    try:
        values = value_str.split(sep)
        if (min_vals is not None and len(values) < min_vals) or (
            max_vals is not None and len(values) > max_vals
        ):
            raise InputException(error_msg + " " + value_str)
        values_out = []
        for value in values:
            value = a_type(value)
            if (lb is not None and value < lb) or (ub is not None and value > ub):
                raise InputException(error_msg + " " + value_str)
            values_out.append(value)
        return values_out
    except ValueError:
        raise InputException(error_msg + " " + value_str)


def seq_range_validator_advice(validator_args):
    """A function for the seq_range_validator that returns a short help string
    based on the supplied args.

    ([type, object, object, bool, str]) -> str
    """

    sep, min_vals, max_vals, a_type, lb, ub, error_msg = validator_args
    k_str = ""
    if min_vals is not None or max_vals is not None:
        k_str = "k"
    if min_vals is not None:
        k_str = str(min_vals) + " <= " + k_str
    if max_vals is not None:
        k_str = k_str + " <= " + str(max_vals)
    x_str = ""
    if lb is not None or ub is not None:
        x_str = "xi"
    if lb is not None:
        x_str = str(lb) + " <= " + x_str
    if ub is not None:
        x_str = x_str + " <= " + str(ub)
    adv_str = ""
    if k_str:
        adv_str += k_str
    if x_str:
        if k_str:
            adv_str += ", "
        adv_str += x_str
    return " {" + adv_str + "}"


def print_usage(arg_processor):
    """This function prints a usage message to sys.stdout based on the
    current argument and flag definition. It then exits.

    (ArgProcessor) -> None
    """
    min_width = max(
        
            len(x)
            for x in arg_processor.program_arg_order + arg_processor.program_flag_order
        
    )
    print(f"Usage: python {sys.argv[0]} ")
    print("The following flags and arguments can be supplied:")
    print("Flags:")
    for flag in arg_processor.program_flag_order:
        print(
            "  {:<{}} : {}".format(
                flag, min_width, arg_processor.program_flags[flag].description
            )
        )
    print("Arguments:")
    for arg in arg_processor.program_arg_order:
        if arg_processor.program_args[arg].validator is not None:
            advice_str = arg_processor.advice_functions[
                arg_processor.program_args[arg].validator
            ](arg_processor.program_args[arg].validator_args)
        else:
            advice_str = ""
        if arg_processor.program_args[arg].needed:
            print(
                "  {:<{}} : {}{}".format(
                    arg,
                    min_width,
                    arg_processor.program_args[arg].description,
                    advice_str,
                )
            )
        else:
            print(
                "  {:<{}} : {}{} [optional, default: {}]".format(
                    arg,
                    min_width,
                    arg_processor.program_args[arg].description,
                    advice_str,
                    arg_processor.program_args[arg].default_value,
                )
            )
    sys.exit(0)


# The main class
# -------------------------------------------------------------------------------


class ArgProcessor:
    """Objects instantiating this class will have a number of argument and
    flag definitions added. Once parse_args() has been called, the
    appropriate local variables of the ArgProcessor object will be set or
    an InputException will be raised.
    """

    def __init__(self):
        """Make a new ArgProcessor.

        (ArgProcessor) -> None
        """
        self.program_args = {}
        self.program_arg_order = []
        self.program_flags = {}
        self.program_flag_order = []
        self.validators = []
        self.advice_functions = {}
        # Register the pre-made validators
        self.register_validator(range_validator, range_validator_advice)
        self.register_validator(seq_range_validator, seq_range_validator_advice)
        self.register_validator(enum_validator, enum_validator_advice)
        self.register_validator(bool_validator, bool_validator_advice)
        # Add the quiet flag:
        self.add_program_flag(
            "--quiet", FlagDefinition("quiet", None, "Suppress non-essential output")
        )
        # Add the help flag
        self.add_program_flag(
            "--help", FlagDefinition("help", print_usage, "Display this notice")
        )

    def add_program_arg(self, param, arg_definition):
        """This method adds a program argument.

        (ArgProcessor, str, ArgDefinition) -> None
        param:          The '-' prefixed param string of the argument.
        arg_definition: The ArgDefinition object describing the argument.
        """
        assert param not in self.program_args, "Error: parameter name in use."
        assert (
            arg_definition.validator is None
            or arg_definition.validator in self.validators
        ), "Error: unregistered validator"
        self.program_args[param] = arg_definition
        self.program_arg_order.append(param)

    # This method adds a program flag.
    # param:          The '--' prefixed flag string.
    # arg_definition: The FlagDefinition object describing the flag.
    def add_program_flag(self, flag, flag_definition):
        assert flag not in self.program_flags, "Error: flag name in use."
        self.program_flags[flag] = flag_definition
        self.program_flag_order.append(flag)
        vars(self)[flag_definition.var_name] = False

    def register_validator(self, validator, validator_advice):
        """This method adds a validator and maps it to the relevant advice
        function.

        (ArgProcessor, function, function) -> None
        """
        self.validators.append(validator)
        self.advice_functions[validator] = validator_advice

    def parse_args(self):
        """Call this function to parse the input arguments.

        (ArgProcessor) -> None
        """
        # -----------------------------------------------------------------------
        # This code is based on code from the KR Toolkit by Christian Muise
        # URL: http://code.google.com/p/krtoolkit/
        try:
            argv, opts, flags = sys.argv[1:], {}, []
            while argv:
                if argv[0][0:2] == "--":
                    flags.append(argv[0])
                    argv = argv[1:]
                elif argv[0][0] == "-":
                    opts[argv[0]] = argv[1]
                    argv = argv[2:]
                else:
                    raise InputException("Badly constructed arg: " + argv[0])
        except IndexError:
            raise InputException("Badly constructed arg: " + argv[0])
        # -----------------------------------------------------------------------
        for flag in flags:
            if flag in self.program_flags:
                vars(self)[self.program_flags[flag].var_name] = True
                if self.program_flags[flag].function:
                    self.program_flags[flag].function(self)
            else:
                raise InputException("Invalid flag: " + flag)

        if not self.quiet:
            min_width = (
                max(
                    len("Flags:"),
                    max(len(x.description) for x in iter(self.program_args.values())),
                )
                + 1
            )
            if len(flags) == 0:
                print("{:<{}} {}".format("Flags:", min_width, "<None>"))
            else:
                print(
                    "{:<{}} {}".format(
                        "Flags:",
                        min_width,
                        ", ".join([f for f in self.program_flags if f in flags]),
                    )
                )

        for arg in opts:
            if arg not in self.program_args:
                raise InputException("Invalid arg: " + arg)

        for arg in self.program_arg_order:
            arg_def = self.program_args[arg]
            if arg not in opts:
                if arg_def.needed:
                    raise InputException("Error needed arg is missing: " + arg)
                vars(self)[arg_def.var_name] = arg_def.default_value
            else:
                if arg_def.validator is None:
                    vars(self)[arg_def.var_name] = opts[arg]
                else:
                    vars(self)[arg_def.var_name] = arg_def.validator(
                        opts[arg], arg_def.validator_args
                    )
            if not self.quiet:
                print(
                    "{:<{}} {}".format(
                        arg_def.description + ":",
                        min_width,
                        vars(self)[arg_def.var_name],
                    )
                )



#! /usr/bin/python3

"""
Wrapper for Pathways.

Creates domain.pddl and problem.pddl files in the current directory.
"""

import argparse
from collections import defaultdict
import os.path
from pathlib import Path
import re
import subprocess
import sys

DIR = Path(__file__).parent.resolve()

DOMAIN_HEADER = """\
(define (domain Pathways-Propositional)
(:requirements :typing)

(:types
    level molecule - object
    simple complex - molecule
)
"""

PREDICATES = """\
(:predicates
    (association-reaction ?x1 ?x2 - molecule ?x3 - complex)
    (catalyzed-association-reaction ?x1 ?x2 - molecule ?x3 - complex)
    (synthesis-reaction ?x1 ?x2 - molecule)
    (possible ?x - molecule)
    (available ?x - molecule)
    (chosen ?s - simple)
    (next ?l1 ?l2 - level)
    (num-subs ?l - level)
{goal_predicates}
)
"""

ACTIONS = """\
(:action choose
 :parameters (?x - simple ?l1 ?l2 - level)
 :precondition (and (possible ?x) (not (chosen ?x))
                    (num-subs ?l2) (next ?l1 ?l2))
 :effect (and (chosen ?x) (not (num-subs ?l2)) (num-subs ?l1)))

(:action initialize
  :parameters (?x - simple)
  :precondition (and (chosen ?x))
  :effect (and (available ?x)))

(:action associate
 :parameters (?x1 ?x2 - molecule ?x3 - complex)
 :precondition (and (association-reaction ?x1  ?x2  ?x3)
                    (available ?x1) (available ?x2))
 :effect (and  (not (available ?x1)) (not (available ?x2)) (available ?x3)))

(:action associate-with-catalyze
 :parameters (?x1 ?x2 - molecule ?x3 - complex)
 :precondition (and (catalyzed-association-reaction ?x1 ?x2 ?x3)
                    (available ?x1) (available ?x2))
 :effect (and (not (available ?x1)) (available ?x3)))

(:action synthesize
 :parameters (?x1 ?x2 - molecule)
 :precondition (and (synthesis-reaction ?x1 ?x2) (available ?x1))
 :effect (and (available ?x2)))
"""


def parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--goals", type=int, default=1)
    parser.add_argument("--reactions", type=int, default=1)
    parser.add_argument("--initial-substances", type=int, default=1)
    parser.add_argument("--seed", type=int, default=2004)
    parser.add_argument("domain")
    parser.add_argument("problem")
    return parser.parse_args()


def get_simplesubs():
    with open(DIR / "Pathways-SimpleSubs") as f:
        return set(f.read().splitlines())


def extract_molecules(problem_file):
    molecules = defaultdict(list)
    new_lines = []
    with open(problem_file) as f:
        for line in f:
            match = re.match(r"\t(.+) - (simple|complex)\n", line)
            if match:
                molecules[match.group(2)].append(match.group(1))
            else:
                new_lines.append(line)
    with open(problem_file, "w") as f:
        f.writelines(new_lines)
    return molecules


def get_goals(dummy_actions):
    molecules = re.findall(r"\(available ([^\)]+)\)", dummy_actions)
    goals = re.findall(r":effect \(and \((.+)\)\)\)", dummy_actions)
    assert len(molecules) % 2 == 0
    assert len(goals) == len(molecules) / 2
    return list(zip(molecules[::2], molecules[1::2], goals))


def make_strips_dummy_action(molecule, action_index, goal):
    return f"""\
(:action DUMMY-STRIPS-ACTION-{action_index}
 :parameters ()
 :precondition (available {molecule})
 :effect (and ({goal})))
"""


def main():
    args = parse_args()

    simplesubs = get_simplesubs()

    problem_file = os.path.abspath(args.problem)

    p = subprocess.run(
        [
            os.path.join(DIR, "pathways"),
            "--seed",
            str(args.seed),
            "-out",
            problem_file,
            "-R",
            str(args.reactions),
            "-G",
            str(args.goals),
            "-L",
            str(args.initial_substances),
            "-n",
            "Pathways-Problem",
        ],
        stdout=subprocess.PIPE,
        universal_newlines=True,
        check=True,
    )
    dummy_actions_string_original = p.stdout.strip().replace("\t", "    ")

    # Define molecules in domain file instead of problem file. Not sure if this is necessary.
    molecules = extract_molecules(problem_file)
    duplicate_molecules = set(molecules["simple"]) & set(molecules["complex"])
    if duplicate_molecules:
        sys.exit(f"Error: simple and complex molecules overlap: {duplicate_molecules}")

    goals = get_goals(dummy_actions_string_original)

    for mol1, mol2, _goal in goals:
        for molecule in [mol1, mol2]:
            if molecule not in set(molecules["simple"]) | set(molecules["complex"]):
                print(f"Molecule {molecule} is undefined --> will define it now")
                if molecule in simplesubs:
                    molecules["simple"].append(molecule)
                else:
                    molecules["complex"].append(molecule)

    # Turn each disjunctive dummy action into two STRIPS actions with a single precondition.
    dummy_strips_actions = []
    for mol1, mol2, goal in goals:
        for molecule in [mol1, mol2]:
            dummy_strips_actions.append(
                make_strips_dummy_action(molecule, len(dummy_strips_actions), goal)
            )

    constants_pddl = "(:constants\n    {} - simple\n\n    {} - complex)\n".format(
        " ".join(molecules["simple"]), " ".join(molecules["complex"])
    )

    goal_predicates = "\n".join(
        [f"    (goal{goal})" for goal in range(1, args.goals + 1)]
    )
    dummy_actions_string = "\n".join([f"{action}" for action in dummy_strips_actions])

    domain_parts = [
        DOMAIN_HEADER,
        constants_pddl,
        PREDICATES.format(**locals()),
        ACTIONS,
        f"{dummy_actions_string}\n)",
        "",
    ]

    with open(args.domain, "w") as f:
        f.write("\n".join(domain_parts))


main()



import json
from importlib import resources
from pprint import pprint
from typing import List, Dict, Sequence


def get_domain_names() -> List[str]:
    """
    List all benchmark domain names.
    """
    pkg = f"amlgym.benchmarks.domains"
    return [f.name.split('.')[0] for f in resources.files(pkg).iterdir() if f.is_file()]


def print_domains() -> None:
    """
    Print all benchmark domain names.
    """
    pprint(get_domain_names())


def get_domain(domain_name: str) -> str:
    """
    Read the content of a PDDL domain file as text.
    """
    pkg = f"amlgym.benchmarks.domains"
    domain_file = f"{domain_name}.pddl" if '.pddl' not in domain_name else domain_name
    with resources.open_text(pkg, domain_file) as f:
        return f.read()


def get_trajectories(domain_name: str,
                     kind: str = 'learning') -> List[str]:
    """
    Return a list of trajectory strings for a PDDL domain in the benchmarks.trajectories package.
    """
    base_pkg = "amlgym.benchmarks.trajectories"

    possible_kinds = sorted([p.name for p in resources.files(base_pkg).iterdir()
                             if p.is_dir() and not p.name.startswith("_")])

    assert kind in possible_kinds, f'`kind` must be in {possible_kinds}'

    pkg = f"{base_pkg}.{kind}.{domain_name.split('.')[0]}"
    trajectories = []
    for traj_file in sorted(resources.files(pkg).iterdir(),
                            key=lambda x: int(x.name.split('_')[0])):
        with resources.open_text(pkg, traj_file.name) as f:
            trajectories.append(f.read())
    return trajectories


def get_domain_path(domain_name: str) -> str:
    """
    Return the absolute path of a PDDL domain file in the benchmarks.domains package.
    """
    pkg = "amlgym.benchmarks.domains"
    domain_name = f"{domain_name}.pddl" if '.pddl' not in domain_name else domain_name
    # get absolute path of pddl domain file
    domain_path = resources.files(pkg).joinpath(domain_name)
    return str(domain_path)


def get_trajectories_path(domain_name: str,
                          kind: str = 'learning') -> List[str]:
    """
    Return the absolute path of a PDDL domain trajectory files in the benchmarks.trajectories package.
    """
    base_pkg = "amlgym.benchmarks.trajectories"

    possible_kinds = sorted([p.name for p in resources.files(base_pkg).iterdir()
                             if p.is_dir() and not p.name.startswith("_")])

    assert kind in possible_kinds, f'`kind` must be in {possible_kinds}'

    pkg = f"{base_pkg}.{kind}.{domain_name.split('.')[0]}"

    trajectories_path = [str(f) for f in resources.files(pkg).iterdir() if f.is_file()]
    return sorted(trajectories_path, key=lambda x: int(x.split('/')[-1].split('_')[0]))


def get_problems_path(domain_name: str,
                      kind: str = 'solving') -> List[str]:
    """
    Return the absolute path of a PDDL domain problem files in the benchmarks.problems package.
    """
    base_pkg = "amlgym.benchmarks.problems"

    possible_kinds = sorted([p.name for p in resources.files(base_pkg).iterdir()
                             if p.is_dir() and not p.name.startswith("_")])

    assert kind in possible_kinds, f'`kind` must be in {possible_kinds}'

    pkg = f"{base_pkg}.{kind}.{domain_name.split('.')[0]}"
    problems_path = [str(f) for f in resources.files(pkg).iterdir() if f.is_file()]
    return sorted(problems_path, key=lambda x: int(x.split('/')[-1].split('_')[0]))


def get_test_states(domain_name: str,
                    kind: str = 'predictive_power') -> Dict[str, Sequence[object]]:
    """
    Return a set of test states from some JSON format for a PDDL domain in the
    benchmarks.states package. The returned set of test states is a dictionary
    where keys are PDDL problem files and values are list of test states.
    """
    base_pkg = "amlgym.benchmarks.states"

    possible_kinds = sorted([p.name for p in resources.files(base_pkg).iterdir()
                             if p.is_dir() and not p.name.startswith("_")])

    assert kind in possible_kinds, f'`kind` must be in {possible_kinds}'

    pkg = f"{base_pkg}.{kind}.{domain_name.split('.')[0]}"

    try:
        states_path = str(next(f for f in resources.files(pkg).iterdir() if f.is_file()))
    except StopIteration:
        raise FileNotFoundError(f"No files found in package {pkg}.")

    with open(states_path, 'r') as f:
        test_set = json.load(f)

    return test_set



#!/usr/bin/env python
# Generator for the child-snack domain
#
# Authors: Raquel Fuentetaja and Tomas de la Rosa
# See MIT License attached
#
# See domain file for description.
# The ratio parameter is the proportion of needed symbols
# that are declared in advance in the problem file. The min ratio
# should be 1.0 to guarantee solvability
#
# The generator is also able to generate problems in alternative
# representations:
#   (1) Pool: The standard pizza domain
#   (2) Control: A modified version with precondition control for
#   creating one slice at a time.
#   (3) Numeric: An equivalent version in numeric PDDL, where
#   no slice symbols have to be declared in advance.
#


from math import ceil
import random
import sys


class Task:
    def __init__(
        self,
        seed,
        n_children,
        n_trays,
        gluten_factor,
        domain_type="pool",
        const_ratio=1.2,
    ):
        self.type = domain_type
        self.seed = seed
        self.n_children = n_children
        self.n_trays = n_trays
        self.gluten_factor = gluten_factor
        self.const_ratio = const_ratio

        self.n_sandwiches = int(ceil(self.n_children * self.const_ratio))

        self.tables = enum_objects("table", 3)

        self.trays = enum_objects("tray", self.n_trays)
        self.children = enum_objects("child", self.n_children)
        self.sandwiches = enum_objects("sandw", self.n_sandwiches)
        self.breads = enum_objects("bread", self.n_children)
        self.contents = enum_objects("content", self.n_children)

        self.n_allergie = int(self.n_children * self.gluten_factor)

    def dump(self, out=None):
        self._dump_header(out)
        self._dump_objects("  ", out)
        self._dump_init("  ", out)
        self._dump_goal("  ", out)
        # self._dump_metric("  ", out)
        print(")", file=out)

    def _dump_header(self, out=None):
        print(
            "; child-snack task with %s children and %s gluten factor "
            % (self.n_children, self.gluten_factor),
            file=out,
        )
        print("; constant factor of %s" % self.const_ratio, file=out)
        print("; random seed: %d" % self.seed, file=out)
        print(file=out)
        print("(define (problem prob-snack)", file=out)
        print("  (:domain child-snack)", file=out)

    def _dump_objects(self, indent="", out=None):
        print(indent + "(:objects", file=out)
        print("{}  {} - child".format(indent, " ".join(self.children)), file=out)
        print("{}  {} - bread-portion".format(indent, " ".join(self.breads)), file=out)
        print("{}  {} - content-portion".format(indent, " ".join(self.contents)), file=out)
        print("{}  {} - tray".format(indent, " ".join(self.trays)), file=out)
        print("{}  {} - place".format(indent, " ".join(self.tables)), file=out)

        if self.type != "numeric":
            print("{}  {} - sandwich".format(indent, " ".join(self.sandwiches)), file=out)

        print(indent + ")", file=out)

    def _dump_init(self, indent="", out=None):
        print(indent + "(:init", file=out)

        for t in self.trays:
            print(indent + f"{indent} (at {t} kitchen)", file=out)
        for b in self.breads:
            print(indent + f"{indent} (at_kitchen_bread {b})", file=out)
        for c in self.contents:
            print(indent + f"{indent} (at_kitchen_content {c})", file=out)

        for b in random.sample(self.breads, self.n_allergie):
            print(indent + f"{indent} (no_gluten_bread {b})", file=out)

        for c in random.sample(self.contents, self.n_allergie):
            print(indent + f"{indent} (no_gluten_content {c})", file=out)

        allergie_children = set(random.sample(self.children, self.n_allergie))
        no_allergie_children = set(self.children).difference(allergie_children)

        for c in list(allergie_children):
            print(indent + f"{indent} (allergic_gluten {c})", file=out)
        for c in list(no_allergie_children):
            print(indent + f"{indent} (not_allergic_gluten {c})", file=out)

        for c in self.children:
            print(
                indent + f"{indent} (waiting {c} {random.choice(self.tables)})",
                file=out,
            )

        if self.type == "pool":
            for s in self.sandwiches:
                print(indent + f"{indent} (notexist {s})", file=out)

        elif self.type == "control":
            print(f"{indent}   (current {self.sandwiches[0]})", file=out)
            for i in range(1, len(self.sandwiches)):
                print(f"{indent}   (next sandw{i} sandw{i + 1})", file=out)
        elif self.type == "numeric":
            print("%s   (= (nsandwiches-at_kitchen-no_gluten)  0)" % indent, file=out)
            print("%s   (= (nsandwiches-at_kitchen)  0)" % indent, file=out)
            for t in self.trays:
                print(
                    f"{indent}   (= (nsandwiches-ontray-no_gluten {t})  0)",
                    file=out,
                )
                print(f"{indent}   (= (nsandwiches-ontray {t})  0)", file=out)

        print(indent + ")", file=out)

    def _dump_goal(self, indent="", out=None):
        print(indent + "(:goal", file=out)
        print(indent + "  (and", file=out)
        for g in self.children:
            print(indent + f"{indent} (served {g})", file=out)
        print(indent + "  )", file=out)
        print(indent + ")", file=out)

    def _dump_metric(self, indent="", out=None):
        print(indent + "(:metric minimize (total-cost))", file=out)


def enum_objects(pddltype, n):
    return [f"{pddltype}{i}" for i in range(1, n + 1)]


def usage():
    print("\n Incorrect input arguments ")
    print(
        "\n Usage: "
        + sys.argv[0]
        + " [pool | control | numeric]"
        + " <seed> <num_children> <num_trays> <gluten_factor> <const_ratio>"
    )
    print("\n    <const_ratio>: min 1.0")
    sys.exit(1)


if __name__ == "__main__":

    if len(sys.argv) < 6:
        usage()

    try:
        domain_type = sys.argv[1]
        seed = int(sys.argv[2])
        n_children = int(sys.argv[3])
        n_trays = int(sys.argv[4])
        gluten_factor = float(sys.argv[5])
        const_ratio = float(sys.argv[6])

        random.seed(seed)
        pddltask = Task(
            seed, n_children, n_trays, gluten_factor, domain_type, const_ratio
        )
        pddltask.dump()

    except BaseException:
        usage()



import os
import itertools
import shutil
from collections import defaultdict
from unified_planning.model import Fluent
import re
from typing import List, Dict, Set, ClassVar, OrderedDict
from amlgym.algorithms.OfflineAlgorithmAdapter import OfflineAlgorithmAdapter
from unified_planning.io import PDDLReader
from offlam.algorithm import learn


class OffLAM(OfflineAlgorithmAdapter):
    """
    Adapter class for running the OffLAM algorithm: "Lifted Action Models Learning
    from Partial Traces", L. Lamanna, L. Serafini, A. Saetti, A. Gerevini,
    and P. Traverso, Artificial Intelligence Journal, 2025.
    https://www.sciencedirect.com/science/article/abs/pii/S0004370224001929

    Example:
        .. code-block:: python

            from amlgym.algorithms import get_algorithm
            offlam = get_algorithm('OffLAM')
            model = offlam.learn('path/to/domain.pddl', ['path/to/trace0', 'path/to/trace1'])
            print(model)

    """
    _reference: ClassVar[OrderedDict[str, str]] = {
        'Authors': "L. Lamanna, L. Serafini, A. Saetti, A. Gerevini, and P. Traverso",
        'Title': "Lifted Action Models Learning from Partial Traces",
        'Venue': "Artificial Intelligence Journal",
        'Year': 2025,
        'URL': "https://www.sciencedirect.com/science/article/abs/pii/S0004370224001929"
    }

    def __init__(self, **kwargs):
        super(OffLAM, self).__init__(**kwargs)

    def learn(self,
              domain_path: str,
              trajectory_paths: List[str]) -> str:

        # Fill input trajectories with some (i.e. `relevant`) missing literals
        if os.path.exists('tmp'):
            if not os.path.isdir("tmp"):
                os.remove("tmp")
            else:
                shutil.rmtree('tmp')
        os.makedirs('tmp', exist_ok=True)
        filled_traj_paths = []
        for i, traj_path in enumerate(trajectory_paths):
            filled_traj = self._preprocess_trace(domain_path, traj_path)  # add relevant negative literals
            filled_traj_paths.append(f"tmp/{i}_traj_filled")
            with open(f"tmp/{i}_traj_filled", "w") as f:
                f.write(filled_traj)

        # Learn action model
        model = learn(domain_path, filled_traj_paths)

        # TODO: open issue in OffLAM
        model = model.replace("(:requirements)", "(:requirements :typing)")

        # Remove temporary files
        shutil.rmtree('tmp')

        return model

    def _preprocess_trace(self, domain_path: str, traj_path: str) -> str:
        """
        Format the trajectory to make it compliant with the algorithm, by explicitly
        stating negative literals.

        :parameter domain_path: path to the input domain file
        :parameter traj_path: path to the trajectory file

        :return: a string representing the formatted trajectory
        """

        # Inner helper function
        def ground_atoms(atom: Fluent,
                         objects: Dict[str, Set[str]]) -> Set[str]:
            """
            Ground a lifted atom with a set of objects by checking object types are in the atom signature
            :param atom: a lifted atom
            :param objects: dictionary where keys are object ids and values object types
            :return: list of grounded atoms
            """
            atom_objs = [[o for o, o_types in objects.items() if param.type.name in o_types]
                         for param in atom.signature]
            if len(atom_objs) == 0:
                return {f"({atom.name})"}
            return {f"({atom.name} {' '.join(comb)})" for comb in itertools.product(*atom_objs)}

        domain = PDDLReader().parse_problem(domain_path)
        with open(traj_path, 'r') as f:
            traj_str = f.read()
            traj_str = re.sub(r' +', ' ', traj_str)  # format extra spaces

        states = [r for r in traj_str.split('\n') if r.strip().startswith('(:state ')]
        actions = [{'name': a.split()[0], 'objs': a.split()[1:] if len(a.split()) > 1 else list()}
                   for a in re.findall(r"\(:action\s+\((.*?)\)\)", traj_str)]
        states = [{
            'pos': {e.strip() for e in re.findall(r"\([^()]*\)", s)
                    if not len(e.replace('(and', '').replace(')', '').strip()) == 0},
            'neg': set()}
            for s in states]

        # for every object, get all types compatible with the predicate signature of observable atoms
        objects_types = defaultdict(set)

        for i in range(len(states)):
            s = states[i]
            for l in s['pos']:
                terms = l.strip()[1:-1].split()
                pred = terms[0]
                objs = list(terms[1:]) if len(terms) > 1 else list()
                for k, o in enumerate(objs):
                    objects_types[o].add(domain.fluent(pred).signature[k].type.name)

        for i in range(len(states) - 1):
            s = states[i]

            # Compute all literals involving next action objects
            action_objs = {o: objects_types[o] for o in actions[i]['objs']}

            relevant_literals = set()
            for atom in domain.fluents:
                # relevant_literals = relevant_literals.union(ground_atoms(atom, objects_types))
                relevant_literals = relevant_literals.union(ground_atoms(atom, action_objs))

            # Add missing negative literals to prev state
            neg = relevant_literals - s['pos']
            s['neg'] = s['neg'].union({f"(not {l})" for l in neg})

            # Add missing negative literals to next state
            neg = relevant_literals - states[i+1]['pos']
            states[i+1]['neg'] = states[i+1]['neg'].union({f"(not {l})" for l in neg})

        traj_str = "(:observation "
        for i in range(len(states) - 1):
            traj_str += f"\n\n(:state {' '.join(states[i]['pos'].union(states[i]['neg']))})"
            if len(actions[i]['objs']) > 0:
                traj_str += f"\n\n(:action ({actions[i]['name']} {' '.join(actions[i]['objs'])}))"
            else:
                traj_str += f"\n\n(:action ({actions[i]['name']}))"

        traj_str += f"\n\n(:state {' '.join(states[-1]['pos'].union(states[-1]['neg']))})\n\n)"

        return traj_str



r"""
Predictive power metrics assess a domain model ability of predicting actions
applicability and outcomes of executing an action in a state.
The *predicted applicability* evaluates the model ability of predicting whether 
and action is applicable in an environment state.
The *predicted effects* evaluates the model ability of predicting the state reached 
after executing an action in an environment state.
Predictive power metrics are defined with respect to a set of (test) states and a simulator, 
which is required to evaluate if an action is applicable and what is the state reached 
after executing the action in a given (simulated) environment state.
"""
import logging
import warnings
from collections import defaultdict
from contextlib import nullcontext

import unified_planning.model
from alive_progress import alive_bar
from typing import Dict, Sequence
import numpy as np
from unified_planning.shortcuts import SequentialSimulator

from amlgym.modeling.env import Env, StateType, ActionType

# Disable printing of planning engine credits to avoid overloading stdout
unified_planning.shortcuts.get_environment().credits_stream = None


def applicability(simulator: Env | Sequence[Env],
                  simulator_ref: Env | Sequence[Env],
                  test_states: Sequence[StateType] | Sequence[Sequence[StateType]],
                  applicable_actions: Sequence[ActionType] = None,
                  show_progress: bool = True) -> Dict[str, Dict[str, float]]:
    r"""
    Evaluate the predicted applicability metric given the simulator of a domain model :math:`M` and an
    environment simulator :math:`E`. The model :math:`M` and environment simulators share the set :math:`S`
    of states and the set :math:`A` of actions.
    For an action :math:`a\in A`, and state :math:`s\in S`,
    we denote by :math:`app_M(a,S_{test})` and :math:`app(a,S_{test})` the set of states in :math:`S_{test}`
    in which :math:`a` is applicable according to :math:`M` and :math:`E`, respectively.
    We define the predicted applicability metric for every action :math:`a\in A` as:

    * True Positives: :math:`TP_{app}(a)=|app_M(a,S_{test})\cap app(a,S_{test})|`
    * False Positives: :math:`FP_{app}(a)=|app_M(a,S_{test})\setminus app(a,S_{test})|`
    * False Negatives: :math:`FN_{app}(a)=|app(a,S_{test}) \setminus app_M(a,S_{test})|`

    The predicted applicability precision and recall per action are obtained as:

    * Predicted applicability precision of :math:`a`: :math:`P_{app}(a) = \frac{TP_{app}(a)}{TP_{app}(a)+FP_{app}(a)}`
    * Predicted applicability recall of :math:`a`: :math:`R_{app}(a) = \frac{TP_{app}(a)}{TP_{app}(a)+FN_{app}(a)}`

    When :math:`TP_{app}(a) = FP_{app}(a) = 0`, we define :math:`P_{app}(a)=1` and :math:`R_{app}(a)=0`, as the
    domain model :math:`M` never allows :math:`a` to be applied in :math:`S_{test}`.
    Finally, the mean precision and recall of a domain model averages over the actions precision and recall:

    * **Predicted applicability precision** of :math:`M`: :math:`P = \frac{1}{|A|}\sum_{a\in A} P(a)`

    * **Predicted applicability recall** of :math:`M`: :math:`R = \frac{1}{|A|}\sum_{a\in A} R(a)`

    :param simulator: simulator of a domain model :math:`M` to be evaluated.
    :param simulator_env: environment simulator :math:`E` to compare with.
    :param test_states: set :math:`S_{test}` of test states.
    :param applicable_actions: optionally precomputed set of applicable actions for every test state.
    :param show_progress: show a progress bar.

    :return: the predicted applicability precision and recall averaged over all test states and actions.
    """

    if isinstance(simulator, Sequence):
        assert isinstance(simulator_ref, Sequence), ('The applicability evaluation for a sequence of '
                                                     'learned simulators requires a corresponding sequence '
                                                     'of reference simulators: a reference simulator to compare with '
                                                     'for each learned simulator.')
        assert len(simulator_ref) == len(simulator), ("The number of learned and reference simulators "
                                                              "must be equal. Current number of learned simulators "
                                                              f"is {len(simulator)} and reference simulators "
                                                              f"is {len(simulator_ref)}.")
        assert len(simulator_ref) == len(test_states), ("A sequence of states to be evaluated must be provided "
                                                        "for every pair (learned, reference) simulator. "
                                                        f"The current sequence of states has size {len(test_states)}, "
                                                        f"while the number of (learned, reference) simulator "
                                                        f"pairs is {len(simulator_ref)}.")
        simulator_ref_list = simulator_ref
        simulator_learned_list = simulator
        test_states_list = test_states
    else:
        simulator_ref_list = [simulator_ref]
        simulator_learned_list = [simulator]
        test_states_list = [test_states]

    tp = defaultdict(int)
    fp = defaultdict(int)
    fn = defaultdict(int)
    precision = defaultdict(float)
    recall = defaultdict(float)

    bar = alive_bar(len(test_states_list),
                    title=f'Evaluating actions applicability...',
                    length=20) if show_progress else nullcontext()
    with bar as bar:
        for k, (simulator, simulator_ref, states) in enumerate(zip(simulator_learned_list,
                                                                   simulator_ref_list,
                                                                   test_states_list)):
            applicable_in_states = None

            if applicable_actions is not None and k < len(applicable_actions):
                # applicable_actions is a possibly precomputed list of applicable
                # actions for every k-th state in the test set
                applicable_in_states = applicable_actions[k]

            for j, s in enumerate(states):

                s = set(s)  # TODO: use sets in yaml rather than JSON

                # parse the possibly precomputed list of applicable action for the k-th state
                if applicable_in_states is not None:
                    applicable_ref = applicable_in_states[j]
                else:
                    # if no precomputed list of applicable actions is given, compute them
                    applicable_ref = simulator_ref.applicable_actions(s)

                applicable_learned = simulator.applicable_actions(s)

                operators = set(list(applicable_learned.keys()) + list(applicable_ref.keys()))

                for op in operators:
                    tp[op] += len(applicable_ref[op] & applicable_learned[op])
                    fp[op] += len(applicable_learned[op] - applicable_ref[op])
                    fn[op] += len(applicable_ref[op] - applicable_learned[op])

            if show_progress:
                bar()

    operators = set(tp.keys()) | set(fp.keys()) | set(fn.keys())
    for op in operators:

        if (tp[op] + fp[op]) == 0:
            warnings.warn(f"No true and false positives for operator {op}, "
                          f"predicted applicability precision set to 1.", stacklevel=2)
            precision[op] = 1.
        else:
            precision[op] = tp[op] / (tp[op] + fp[op])

        if (tp[op] + fn[op]) == 0:
            warnings.warn(f"No true positives and false negatives for operator {op}, "
                          f"predicted applicability recall set to 1.", stacklevel=2)
            recall[op] = 1.
        else:
            recall[op] = tp[op] / (tp[op] + fn[op])

    return {
        'mean_precision': np.mean(list(precision.values())),
        'mean_recall': np.mean(list(recall.values())),
    }


def predicted_effects(simulator: Env | Sequence[Env],
                      simulator_env: Env | Sequence[Env],
                      test_states: Sequence[StateType] | Sequence[Sequence[StateType]],
                      applicable_actions: Sequence[ActionType] = None,
                      show_progress: bool = True) -> Dict[str, Dict[str, float]]:
    r"""
    Evaluate the predicted effects metric given the simulator of a domain model :math:`M` and an
    environment simulator :math:`E`. The model :math:`M` and environment simulators share the set :math:`S`
    of states and the set :math:`A` of actions; the evaluation considers actions applicable in a
    state for both :math:`M` and :math:`E`.
    For an action :math:`a\in A`, and state :math:`s\in S`,
    we denote by :math:`a_{M}(s)` and :math:`a(s)` the state resulting from applying :math:`a` in :math:`s`
    according to :math:`M` and :math:`E`, respectively.
    We define the predicted effect metrics for every state :math:`s \in S_{test}`
    and action :math:`a\in A` as:

    * True Positives: :math:`TP_{eff}(s,a)=|(a_M(s)\setminus s)\cap (a(s)\setminus s)|`
    * False Positives: :math:`FP_{eff}(s,a)=|(a_M(s)\setminus s)\setminus a(s)|`
    * False Negatives: :math:`FN_{eff}(s,a)=|(a_M(s)\cap s)\setminus a(s)|`

    The predicted effects mean precision and recall per action are obtained by averaging over all
    states in :math:`S_{test}`, i.e.:

    * True Positives: :math:`TP_{eff}(a)=\sum\limits_{s\in S_{test}}TP_{eff}(s,a)`
    * False Positives: :math:`FP_{eff}(a)=\sum\limits_{s\in S_{test}}TP_{eff}(s,a)`
    * False Negatives: :math:`FN_{eff}(a)=\sum\limits_{s\in S_{test}}FN_{eff}(s,a)`

    * Predicted effects precision of :math:`a` : :math:`P_{eff}(a) = \frac{TP_{eff}(a)}{TP_{eff}(a)+FP_{eff}(a)}`
    * Predicted effects recall of :math:`a` : :math:`R_{eff}(a) = \frac{TP_{eff}(a)}{TP_{eff}(a)+FN_{eff}(a)}`

    When :math:`TP_{eff}(a) = FP_{eff}(a) = 0`, we define :math:`P_{eff}(a)=1` and :math:`R_{eff}(a)=0`.
    Finally, the mean precision and recall of a domain model averages over the actions precision and recall:

    * **Predicted effects precision** of :math:`M`: :math:`P = \frac{1}{|A|}\sum_{a\in A} P(a)`
    * **Predicted effects recall** of :math:`M`: :math:`R = \frac{1}{|A|}\sum_{a\in A} R(a)`

    :param simulator: simulator of a domain model :math:`M` to be evaluated.
    :param simulator_env: environment simulator :math:`E` to compare with.
    :param test_states: set :math:`S_{test}` of test states.
    :param applicable_actions: optionally precomputed set of applicable actions for every test state.
    :param show_progress: show a progress bar.

    :return: the predicted effects mean precision and recall
    """

    if isinstance(simulator, Sequence):
        assert isinstance(simulator_env, Sequence), ('The applicability evaluation for a sequence of '
                                                     'learned simulators requires a corresponding sequence '
                                                     'of reference simulators: a reference simulator to compare with '
                                                     'for each learned simulator.')
        assert len(simulator_env) == len(simulator), ("The number of learned and reference simulators "
                                                              "must be equal. Current number of learned simulators "
                                                              f"is {len(simulator)} and reference simulators "
                                                              f"is {len(simulator_env)}.")
        assert len(simulator_env) == len(test_states), ("A sequence of states to be evaluated must be provided "
                                                        "for every pair (learned, reference) simulator. "
                                                        f"The current sequence of states has size {len(test_states)}, "
                                                        f"while the number of (learned, reference) simulator "
                                                        f"pairs is {len(simulator_env)}.")
        simulator_ref_list = simulator_env
        simulator_learned_list = simulator
        test_states_list = test_states
    else:
        simulator_ref_list = [simulator_env]
        simulator_learned_list = [simulator]
        test_states_list = [test_states]

    tp = defaultdict(int)
    fp = defaultdict(int)
    fn = defaultdict(int)
    precision = defaultdict(float)
    recall = defaultdict(float)

    bar = alive_bar(len(test_states_list),
                    title=f'Evaluating actions applicability...',
                    length=20) if show_progress else nullcontext()
    with bar as bar:
        for k, (simulator, simulator_env, states) in enumerate(zip(simulator_learned_list,
                                                                   simulator_ref_list,
                                                                   test_states_list)):
            applicable_in_states = None

            if applicable_actions is not None and k < len(applicable_actions):
                # applicable_actions is a possibly precomputed list of applicable
                # actions for every k-th state in the test set
                applicable_in_states = applicable_actions[k]

            for j, s in enumerate(states):

                s = set(s)  # TODO: use sets in yaml rather than JSON

                # parse the possibly precomputed list of applicable action for the k-th state
                if applicable_in_states is not None:
                    applicable_ref = applicable_in_states[j]
                else:
                    # if no precomputed list of applicable actions is given, compute them
                    applicable_ref = simulator_env.applicable_actions(s)

                applicable_learned = simulator.applicable_actions(s)

                operators = set(list(applicable_learned.keys()) + list(applicable_ref.keys()))

                # NOTE: consider only actions that are applicable in a given state
                # according to both the evaluated and reference simulators
                for op in operators:
                    for action_label in applicable_ref[op] & applicable_learned[op]:

                        snext_learned = simulator.apply(s, action_label)
                        snext_ref = simulator_env.apply(s, action_label)

                        tp[op] += len((snext_learned - s) .intersection((snext_ref - s)))
                        fp[op] += len((snext_learned - s) - snext_ref)
                        fn[op] += len((snext_learned.intersection(s)) - snext_ref)

            if show_progress:
                bar()

    operators = set(tp.keys()) | set(fp.keys()) | set(fn.keys())
    for op in operators:

        if (tp[op] + fp[op]) == 0:
            warnings.warn(f"No true and false positives for operator {op}, "
                          f"predicted effects precision set to 1.", stacklevel=2)
            precision[op] = 1.
        else:
            precision[op] = tp[op] / (tp[op] + fp[op])


        if (tp[op] + fn[op]) == 0:
            warnings.warn(f"No true positives and false negatives for operator {op}, "
                          f"predicted effects recall set to 1.", stacklevel=2)
            recall[op] = 1.
        else:
            recall[op] = tp[op] / (tp[op] + fn[op])

    return {
        'mean_precision': np.mean(list(precision.values())),
        'mean_recall': np.mean(list(recall.values())),
    }


def predictive_power(simulator_learned: Env | Sequence[Env],
                     simulator_ref: Env | Sequence[Env],
                     test_states: Sequence[StateType] | Sequence[Sequence[StateType]],
                     applicable_actions: Sequence[ActionType] = None,
                     show_progress: bool = True) -> Dict[str, Dict[str, Dict[str, float]]]:
    """
    Evaluate both the predicted applicability and predicted effects metrics of a simulated
    domain model :math:`M` with respect to an environment simulator :math:`E` against a test
    set of states :math:`S_{test}`.
    The results can be reproduced by separately executing functions
    :func:`~amlgym.metrics._predictive.predicted_effects` and
    :func:`~amlgym.metrics._predictive.applicability`;
    this function performs a joint and more efficient evaluation of both metrics.

    :param simulator: simulator of a domain model :math:`M` to be evaluated.
    :param simulator_env: environment simulator :math:`E` to compare with.
    :param test_states: set :math:`S_{test}` of test states.
    :param applicable_actions: optionally precomputed set of applicable actions for every test state.
    :param show_progress: show a progress bar.

    :return: the mean precision and recall for both predicted effects and predicted applicability
    """
    
    if isinstance(simulator_learned, Sequence):
        assert isinstance(simulator_ref, Sequence), ('The applicability evaluation for a sequence of '
                                                     'learned simulators requires a corresponding sequence '
                                                     'of reference simulators: a reference simulator to compare with '
                                                     'for each learned simulator.')
        assert len(simulator_ref) == len(simulator_learned), ("The number of learned and reference simulators "
                                                              "must be equal. Current number of learned simulators "
                                                              f"is {len(simulator_learned)} and reference simulators "
                                                              f"is {len(simulator_ref)}.")
        assert len(simulator_ref) == len(test_states), ("A sequence of states to be evaluated must be provided "
                                                        "for every pair (learned, reference) simulator. "
                                                        f"The current sequence of states has size {len(test_states)}, "
                                                        f"while the number of (learned, reference) simulator "
                                                        f"pairs is {len(simulator_ref)}.")
        simulator_ref_list = simulator_ref
        simulator_learned_list = simulator_learned
        test_states_list = test_states
    else:
        simulator_ref_list = [simulator_ref]
        simulator_learned_list = [simulator_learned]
        test_states_list = [test_states]

    predeffs_tp = defaultdict(int)
    predeffs_fp = defaultdict(int)
    predeffs_fn = defaultdict(int)
    predeffs_precision = defaultdict(float)
    predeffs_recall = defaultdict(float)

    app_tp = defaultdict(int)
    app_fp = defaultdict(int)
    app_fn = defaultdict(int)
    app_precision = defaultdict(float)
    app_recall = defaultdict(float)

    bar = alive_bar(len(test_states_list),
                    title=f'Evaluating predictive power...',
                    length=20) if show_progress else nullcontext()
    with bar as bar:
        for k, (simulator_learned, simulator_ref, states) in enumerate(zip(simulator_learned_list,
                                                                           simulator_ref_list,
                                                                           test_states_list)):
            applicable_in_states = None

            if applicable_actions is not None and k < len(applicable_actions):
                # applicable_actions is a possibly precomputed list of applicable
                # actions for every k-th states list in the test set
                applicable_in_states = applicable_actions[k]

            for j, s in enumerate(states):

                s = set(s)  # TODO: use sets in yaml rather than JSON

                # parse the possibly precomputed list of applicable action for the j-th state
                # in the k-th states list
                if applicable_in_states is not None:
                    applicable_ref = applicable_in_states[j]
                else:
                    # if no precomputed list of applicable actions is given, compute them
                    applicable_ref = simulator_ref.applicable_actions(s)

                applicable_learned = simulator_learned.applicable_actions(s)

                operators = set(list(applicable_learned.keys()) + list(applicable_ref.keys()))

                # Evaluate action applicability and predicted effects
                for op in operators:

                    # Predicted effects
                    for action_label in applicable_ref[op] & applicable_learned[op]:

                        snext_learned = simulator_learned.apply(s, action_label)
                        snext_ref = simulator_ref.apply(s, action_label)

                        predeffs_tp[op] += len((snext_learned - s) & (snext_ref - s))
                        predeffs_fp[op] += len((snext_learned - s) - snext_ref)
                        predeffs_fn[op] += len((snext_learned & s) - snext_ref)
                        pass

                    # Action applicability
                    app_tp[op] += len(applicable_ref[op] & applicable_learned[op])
                    app_fp[op] += len(applicable_learned[op] - applicable_ref[op])
                    app_fn[op] += len(applicable_ref[op] - applicable_learned[op])

            if show_progress:
                bar()

    operators = set(predeffs_tp.keys()) | set(predeffs_fp.keys()) | set(predeffs_fn.keys())
    for op in operators:

        if (app_tp[op] + app_fp[op]) == 0:
            warnings.warn(f"No true and false positives for operator {op}, "
                          f"predicted applicability precision set to 1.", stacklevel=2)
            app_precision[op] = 1.
        else:
            app_precision[op] = app_tp[op] / (app_tp[op] + app_fp[op])

        if (app_tp[op] + app_fn[op]) == 0:
            warnings.warn(f"No true positives and false negatives for operator {op}, "
                          f"predicted applicability recall set to 1.", stacklevel=2)
            app_recall[op] = 1.
        else:
            app_recall[op] = app_tp[op] / (app_tp[op] + app_fn[op])

        if (predeffs_tp[op] + predeffs_fp[op]) == 0:
            warnings.warn(f"No true and false positives for operator {op}, "
                          f"predicted effects precision set to 1.", stacklevel=2)
            predeffs_precision[op] = 1.
        else:
            predeffs_precision[op] = predeffs_tp[op] / (predeffs_tp[op] + predeffs_fp[op])


        if (predeffs_tp[op] + predeffs_fn[op]) == 0:
            warnings.warn(f"No true positives and false negatives for operator {op}, "
                          f"predicted effects recall set to 1.", stacklevel=2)
            predeffs_recall[op] = 1.
        else:
            predeffs_recall[op] = predeffs_tp[op] / (predeffs_tp[op] + predeffs_fn[op])

    return {
        'applicability': {
            'mean_precision': np.mean(list(app_precision.values())),
            'mean_recall': np.mean(list(app_recall.values())),
        },
        'predicted_effects': {
            'mean_precision': np.mean(list(predeffs_precision.values())),
            'mean_recall': np.mean(list(predeffs_recall.values())),
        }
    }



#!/usr/bin/env python


from math import ceil
import random

_colours = ["black", "white", "red", "green", "blue", "mauve"]
_woods = ["oak", "pine", "cherry", "teak", "mahogany", "beech", "walnut"]
_default_machines = {
    "highspeed-saw": 1,
    "glazer": 1,
    "grinder": 1,
    "immersion-varnisher": 1,
    "planer": 1,
    "saw": 1,
    "spray-varnisher": 1,
}
_only_saws = {
    "highspeed-saw": 1,
    "glazer": 0,
    "grinder": 0,
    "immersion-varnisher": 0,
    "planer": 0,
    "saw": 1,
    "spray-varnisher": 0,
}


class Part:
    def __init__(self, index, woods, colours, problemtype=None):
        self.name = "p%d" % index
        self.problemtype = problemtype
        self.generate_random_goal(woods, colours)
        self.generate_random_init(woods, colours)

    def generate_random_goal(self, woods, colours):
        self.goalprops = dict()
        self.goalprops["treatment"] = random.choice(["varnished", "glazed"])
        self.goalprops["surface-condition"] = random.choice(["smooth", "verysmooth"])
        if self.problemtype != "painting":
            colours = colours + ["natural"]
        self.goalprops["colour"] = random.choice(colours)
        self.goalprops["wood"] = random.choice(woods)
        self.generate_goal_selection()

    def generate_goal_selection(self, nr_goals=None):
        if self.problemtype == "sawing":
            nr_goals = random.choice([0, 1])
            self.goalselection = set()
            if nr_goals:
                self.goalselection.add("wood")
        elif self.problemtype == "painting":
            self.goalselection = {"colour"}
        else:
            if not nr_goals:
                nr_goals = random.choice([2, 2, 2, 3, 4])
            self.goalselection = set(random.sample(list(self.goalprops), nr_goals))

    def generate_random_init(self, woods, colours):
        def gen_preprocessing_status():
            self.initprops["treatment"] = random.choice(
                ["varnished", "glazed", "colourfragments"]
            )
            self.initprops["surface-condition"] = random.choice(
                ["rough", "smooth", "verysmooth"]
            )
            self.initprops["wood"] = self.goalprops["wood"]
            poss_colours = set(colours + ["natural"]) - {self.goalprops["colour"]}
            self.initprops["colour"] = random.choice(list(poss_colours))

        self.initprops = dict()
        if self.problemtype == "sawing":
            status = "unused"
        else:
            status = random.choice(["unused"] * 4 + ["available"])
        if status != "unused":
            while True:
                gen_preprocessing_status()
                goal_initially_satisfied = True
                for prop in self.goalselection:
                    if self.initprops[prop] != self.goalprops[prop]:
                        goal_initially_satisfied = False
                        break
                if not goal_initially_satisfied:
                    break
                self.generate_goal_selection(len(self.goalselection))
        self.size = random.randint(5, 15)

    def dump_init(self, indent="", out=None):
        if self.initprops:
            print(f"{indent}(available {self.name})", file=out)
            for prop, val in self.initprops.items():
                print(f"{indent}({prop} {self.name} {val})", file=out)
        else:
            print(f"{indent}(unused {self.name})", file=out)
        print("%s(= (goal-size %s) %d)" % (indent, self.name, self.size), file=out)

    def dump_goal(self, indent="", out=None):
        print(f"{indent}(available {self.name})", file=out)
        for choice in self.goalselection:
            print(
                f"{indent}({choice} {self.name} {self.goalprops[choice]})",
                file=out,
            )


class Board:
    def __init__(self, index, wood, size):
        self.wood = wood
        self.name = "b%d" % index
        self.size = size
        self.surface = random.choice(["rough", "rough", "rough", "smooth"])

    def dump_init(self, indent="", out=None):
        print(f"{indent}(= (board-size {self.name}) {self.size})", file=out)
        print(f"{indent}(wood {self.name} {self.wood})", file=out)
        print(
            f"{indent}(surface-condition {self.name} {self.surface})", file=out
        )
        print(f"{indent}(available {self.name})", file=out)


class Machine:
    def __init__(self, name, type):
        self.name = name
        self.type = type
        self.colours = set()

    def dump_init(self, indent, out=None):
        print(f"{indent}(idle {self.name})", file=out)
        if self.type == "highspeed-saw":
            print(f"{indent}(empty {self.name})", file=out)
        for colour in self.colours:
            print(f"{indent}(has-colour {self.name} {colour})", file=out)


class Task:
    def __init__(self, seed, nr_parts, problemtype=None, **additional_machines):
        self.seed = seed
        self.problemtype = problemtype
        nr_colours = max(int(round(0.7 * nr_parts)), 2)
        self.colours = _colours[:nr_colours]
        nr_woods = max(int(round(0.25 * nr_parts)), 2)
        self.woods = random.sample(_woods, nr_woods)
        self.parts = [
            Part(nr, self.woods, self.colours, problemtype) for nr in range(nr_parts)
        ]
        self._generate_boards()
        self._generate_machines(**additional_machines)
        self._assign_colours_to_machines()

    def _generate_boards(self):
        if self.problemtype == "sawing":
            extra_wood_factor = 1.0
        else:
            extra_wood_factor = 1.2
        quantities = dict()
        for part in self.parts:
            if not part.initprops:
                wood = part.goalprops["wood"]
                quantities.setdefault(wood, []).append(part.size)

        self.boards = []
        counter = 0
        for wood, quantity in quantities.items():
            if not quantity:
                self.woods.remove("wood")
            else:
                random.shuffle(quantity)
                for amounts in zip(
                    quantity[::4],
                    quantity[1::4] + [0],
                    quantity[2::4] + [0],
                    quantity[3::4] + [0],
                ):
                    size = int(ceil(sum(amounts) * extra_wood_factor))
                    self.boards.append(Board(counter, wood, size))
                    counter += 1

    def _generate_machines(self, **changes):
        machines = dict(_default_machines)
        machines.update(changes)
        self.machines = dict()
        for type, number in machines.items():
            if number:
                m = [Machine(f"{type}{nr}", type) for nr in range(number)]
                self.machines[type] = m

    def _assign_colours_to_machines(self):
        necessary = self._determine_necessary_colours()
        spray_varnishers = self.machines.get("spray-varnisher", [])
        immersion_varnishers = self.machines.get("immersion-varnisher", [])
        glazers = self.machines.get("glazer", [])
        for colour in necessary["varnished"]:
            machines = random.sample(
                spray_varnishers, int(ceil(0.6 * len(spray_varnishers)))
            )
            machines += random.sample(
                immersion_varnishers, int(ceil(0.6 * len(immersion_varnishers)))
            )
            for m in machines:
                m.colours.add(colour)
        for colour in necessary["glazed"]:
            for m in random.sample(glazers, int(ceil(0.6 * len(glazers)))):
                m.colours.add(colour)
        for colour in necessary["unspecified"]:
            machines = random.sample(
                spray_varnishers, int(ceil(0.3 * len(spray_varnishers)))
            )
            machines += random.sample(
                immersion_varnishers, int(ceil(0.3 * len(immersion_varnishers)))
            )
            machines += random.sample(glazers, int(ceil(0.3 * len(glazers))))
            for m in machines:
                m.colours.add(colour)
        for type in ["spray-varnisher", "immersion-varnisher", "glazer"]:
            for m in self.machines.get(type, []):
                if not m.colours:
                    m.colours.add(random.choice(list(self.colours)))

    def _determine_necessary_colours(self):
        used_colours = dict()
        used_colours["varnished"] = set()
        used_colours["glazed"] = set()
        used_colours["unspecified"] = set()
        for part in self.parts:
            if "colour" in part.goalselection:
                colour = part.goalprops["colour"]
                if "treatment" in part.goalselection:
                    used_colours[part.goalprops["treatment"]].add(colour)
                else:
                    used_colours["unspecified"].add(colour)
        return used_colours

    def dump(self, out=None):
        self._dump_header(out)
        self._dump_objects("  ", out)
        self._dump_init("  ", out)
        self._dump_goal("  ", out)
        self._dump_metric("  ", out)
        print(")", file=out)

    def _dump_header(self, out=None):
        if self.problemtype:
            print(
                "; woodworking %s task with %s parts"
                % (self.problemtype, len(self.parts)),
                file=out,
            )
        else:
            print("; woodworking task with %s parts" % len(self.parts), file=out)
        print("; Machines:", file=out)
        for type, machines in self.machines.items():
            print(";   %d %s" % (len(machines), type), file=out)
        print("; random seed: %d" % self.seed, file=out)
        print(file=out)
        print("(define (problem wood-prob)", file=out)
        print("  (:domain woodworking)", file=out)

    def _dump_objects(self, indent="", out=None):
        print(indent + "(:objects", file=out)
        for type, machines in self.machines.items():
            print(
                "{}  {} - {}".format(indent, " ".join([m.name for m in machines]), type),
                file=out,
            )
        if self.colours:
            print("{}  {} - acolour".format(indent, " ".join(self.colours)), file=out)
        print("{}  {} - awood".format(indent, " ".join(self.woods)), file=out)
        print(
            "{}  {} - part".format(indent, " ".join([p.name for p in self.parts])), file=out
        )
        print(
            "{}  {} - board".format(indent, " ".join([b.name for b in self.boards])),
            file=out,
        )
        print(indent + ")", file=out)

    def _dump_init(self, indent="", out=None):
        print(indent + "(:init", file=out)
        if self.problemtype != "sawing":
            print(
                indent + "  (grind-treatment-change varnished colourfragments)",
                file=out,
            )
            print(indent + "  (grind-treatment-change glazed untreated)", file=out)
            print(indent + "  (grind-treatment-change untreated untreated)", file=out)
            print(
                indent + "  (grind-treatment-change colourfragments untreated)",
                file=out,
            )
            print(indent + "  (is-smooth smooth)", file=out)
            print(indent + "  (is-smooth verysmooth)", file=out)
        for type in self.machines:
            for m in self.machines[type]:
                m.dump_init(indent + "  ", out)
        for part in self.parts:
            part.dump_init(indent + "  ", out)
        for board in self.boards:
            board.dump_init(indent + "  ", out)
        print(indent + ")", file=out)

    def _dump_goal(self, indent="", out=None):
        print(indent + "(:goal", file=out)
        print(indent + "  (and", file=out)
        for part in self.parts:
            part.dump_goal(indent + "    ", out)
        print(indent + "  )", file=out)
        print(indent + ")", file=out)

    def _dump_metric(self, indent="", out=None):
        print(indent + "(:metric minimize (total-time))", file=out)


def reseed():
    new_seed = random.randrange(10 ** 6)
    random.seed(new_seed)
    return new_seed


def generate_sawing_instance(nr_parts):
    seed = reseed()
    nr_highspeed = nr_parts // 5 + 1
    saws = dict(_only_saws)
    saws["highspeed-saw"] = nr_highspeed
    return Task(seed, nr_parts, "sawing", **saws)


def generate_painting_instance(nr_parts):
    seed = reseed()
    nr_add_painters = int(ceil(random.normalvariate(0.3 * nr_parts, 1)))
    nr_painters = {"spray-varnisher": 1, "immersion-varnisher": 1, "glazer": 1}
    for _ in range(nr_add_painters):
        add_painter = random.choice(
            ["spray-varnisher", "glazer", "glazer", "immersion-varnisher"]
        )
        nr_painters[add_painter] += 1
    return Task(seed, nr_parts, "painting", **nr_painters)


def generate_general_instance(nr_parts):
    seed = reseed()
    nr_add_machines = int(ceil(random.normalvariate(0.4 * nr_parts, 1)))
    machines = dict(_default_machines)
    types = list(machines.keys())
    for _ in range(nr_add_machines):
        type = random.choice(types)
        machines[type] += 1
    return Task(seed, nr_parts, **machines)


if __name__ == "__main__":
    random.seed(2008 * 2008)

    sawing_tasks = [generate_sawing_instance(size) for size in range(15, 27, 3)]
    painting_tasks = [generate_painting_instance(size) for size in range(15, 24, 3)]
    general_tasks = [generate_general_instance(size) for size in range(15, 24, 3)]
    tasks = sawing_tasks + painting_tasks + general_tasks
    for no, task in enumerate(tasks):
        task.dump(out=open("p%02d.pddl" % (no + 11), "w"))



#! /usr/bin/env python

import numpy
import argparse


def gen_board(size_x, size_y, height, num_towers, seed):
    numpy.random.seed(seed)

    board = [[0 for x in range(size_x)] for y in range(size_y)]

    cells = [(x, y) for x in range(size_x) for y in range(size_y)]

    first = True
    for i in numpy.random.choice(range(len(cells)), num_towers, replace=False):
        (x, y) = cells[i]
        col_height = height if first else numpy.random.choice(range(2, height + 1))
        first = False
        board[y][x] = col_height

    board_name = "random_towers_{}x{}_{}_{}_{}.txt".format(
        size_x, size_y, height, num_towers, seed
    )
    f = open("boards/" + board_name, "w")
    for row in board:
        # print (" ".join(map(str, row)))
        f.write(" ".join(map(str, row)) + "\n")
    f.close()

    return board_name


# Board generation procedure that has more control than the one used for IPC'18 about the
# number of blocks in the resulting towers. We introduce an extra parameter, num_blocks.
# The total number_of blocks will be (num_towers-1)*min_height + max_height +
# extra_blocks. The reasoning is that we will have a tower of max_height, the rest of the
# towers with 1 block, and then the extra_blocks are assigned randomly among the other
# towers.
def gen_board_autoscale(size_x, size_y, min_height, max_height, num_towers, rng):

    board = [[0 for x in range(size_x)] for y in range(size_y)]

    cells = [(x, y) for x in range(size_x) for y in range(size_y)]

    first = True
    for i in rng.choice(range(len(cells)), num_towers, replace=False):
        (x, y) = cells[i]
        col_height = (
            max_height if first else int(rng.choice(range(min_height, max_height + 1)))
        )
        first = False
        board[y][x] = col_height

    # board_name = 'random_towers_{}x{}_{}_{}_{}.txt'.format(size_x, size_y, min_height, max_height, num_towers, seed)

    return board
    # f = open('boards/' + board_name, 'w')
    # for row in board:
    #     #print (" ".join(map(str, row)))
    #     f.write(" ".join(map(str, row)) + '\n')
    # f.close()

    # return board_name


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("--size_x", type=int, default=4)
    parser.add_argument("--size_y", type=int, default=4)
    parser.add_argument("--height", type=int, default=4)
    parser.add_argument("--num_towers", type=int, default=4)
    parser.add_argument("--seed", type=int, default=0)

    args = parser.parse_args()

    gen_board(args.size_x, args.size_y, args.height, args.num_towers, args.seed)


if __name__ == "__main__":
    main()



import abc
from dataclasses import dataclass
from typing import Any, Tuple, Dict

from unified_planning.shortcuts import SequentialSimulator

from amlgym.modeling.trajectory import Trajectory


@dataclass
class OnlineAlgorithmAdapter(abc.ABC):
    """
    An abstract class for an online action model learning algorithm, which defines the abstract interface
    that must be implemented by every (subclass) algorithm adapter.
    """

    @abc.abstractmethod
    def learn(self,
              simulator: SequentialSimulator,
              input_domain_path: str,
              seed: int = 123) -> Tuple[str, Trajectory]:
        """
        Learns a PDDL action model from:
         (i)   a simulator of the environment to learn from
         (ii)    a (possibly empty) input model which is required to specify the predicates and operators signature;

        :parameter simulator: environment simulator
        :parameter input_domain_path: input PDDL domain file path
        :parameter seed: random seed for reproducibility

        :return: a string representing the learned PDDL model, and a JSON specification of the trajectory
        """
        raise NotImplementedError



#! /usr/bin/env python


from collections import defaultdict
import itertools
import random
import sys


MAX_EPSILON_ATTEMPTS = 1000
MAX_CONNECTION_ATTEMPTS = 100
MAX_SEED = 10000000


class Point:
    def __init__(self, name, x, y):
        self.name = name
        self.x = x
        self.y = y

    def __lt__(self, other):
        return (self.x, self.y) < (other.x, other.y)

    def __repr__(self):
        return "Point(%s, %d, %d)" % (self.name, self.x, self.y)

    def distance(self, other):
        return ((self.x - other.x) ** 2 + (self.y - other.y) ** 2) ** 0.5

    def round_distance(self, other):
        return int(round(self.distance(other)))


class Graph:
    def __init__(self):
        self.vertices = []
        self.edges = []

    def add_vertex(self, v):
        self.vertices.append(v)

    def add_edge(self, u, v):
        self.edges.append((u, v))
        self.edges.append((v, u))

    def is_connected(self):
        neighbours = defaultdict(list)
        for u, v in self.edges:
            neighbours[u].append(v)
        reached = set()
        queue = self.vertices[:1]
        reached.add(queue[0])
        while queue:
            for succ in neighbours[queue.pop()]:
                if succ not in reached:
                    reached.add(succ)
                    queue.append(succ)
        return len(reached) == len(self.vertices)

    def dump_pddl(self, out=None):
        for v in self.vertices:
            print("(location %s)" % v.name, file=out)
        for u, v in self.edges:
            dist = u.round_distance(v)
            print(f"(connected {u.name} {v.name})", file=out)
            print("(= (distance %s %s) %d)" % (u.name, v.name, dist), file=out)

    def dump_tikz(self, out=None):
        print(r"\documentclass{article}", file=out)
        print(r"\usepackage{tikz}", file=out)
        print(r"\begin{document}", file=out)
        print(r"\begin{tikzpicture}", file=out)
        for v in self.vertices:
            print(
                r"  \node[fill] (%s) at (%.2f, %.2f) {};"
                % (v.name, v.x * 0.01, v.y * 0.01),
                file=out,
            )
        for u, v in self.edges:
            if u < v:
                print(fr"  \draw ({u.name}) -- ({v.name});", file=out)
        print(r"\end{tikzpicture}", file=out)
        print(r"\clearpage", file=out)
        print(r"\begin{verbatim}", file=out)
        self.dump_pddl(out)
        print(r"\end{verbatim}", file=out)
        print(r"\end{document}", file=out)


def find_suitable_point(graph, width, height, epsilon):
    for attempts in itertools.count():
        if attempts == MAX_EPSILON_ATTEMPTS:
            raise ValueError("failed to place vertex: reduce EPSILON")
        name = "loc-%d" % (len(graph.vertices) + 1)
        x = random.randrange(width)
        y = random.randrange(height)
        p = Point(name, x, y)
        if all(p.round_distance(pp) >= epsilon for pp in graph.vertices):
            return p


def generate(num_vert, width, height, connect_distance, epsilon):
    graph = Graph()
    for _ in range(num_vert):
        p = find_suitable_point(graph, width, height, epsilon)
        graph.add_vertex(p)
        for pp in graph.vertices:
            if 0 < p.round_distance(pp) <= connect_distance:
                graph.add_edge(p, pp)
    return graph


def generate_connected(num_vert, width, height, connect_distance, epsilon):
    for attempts in itertools.count():
        if attempts == MAX_CONNECTION_ATTEMPTS:
            raise ValueError("failed to connect graph: increase CONNECT_DISTANCE")
        graph = generate(num_vert, width, height, connect_distance, epsilon)
        if graph.is_connected():
            return graph


def usage():
    raise SystemExit(
        "usage: %s SEED NUM_VERTICES WIDTH HEIGHT "
        "CONNECT_DISTANCE EPSILON" % sys.argv[0]
    )


def parse_args():
    try:
        parameters = list(map(int, sys.argv[1:]))
    except TypeError:
        usage()
    if len(parameters) != 6:
        usage()
    return parameters[0], parameters[1:]


def main(seed, graph_params):
    if not seed:
        seed = random.randrange(MAX_SEED) + 1
    random.seed(seed)
    try:
        g = generate_connected(*graph_params)
    except ValueError as e:
        raise SystemExit(str(e))
    print("% {} {}".format(seed, " ".join(map(str, graph_params))))
    g.dump_tikz()


if __name__ == "__main__":
    main(*parse_args())



import argparse
import os
from pathlib import Path
import random
import shutil

import tarski
from tarski.io import PDDLReader, FstripsWriter
from tarski.model import Model
from tarski.syntax import Constant, land

arg_parser = argparse.ArgumentParser(description="Generate delivery problems")
arg_parser.add_argument("--grid_size_splits", type=str, required=True, help="Comma separated string of grid size splits for train, validation and test. Instances will include sizes up to the specified size, starting from 2 for the training instance. For example, if the argument is '5,7,9', then the training instances will include sizes in the interval [2, 5], validation instances [6, 7] and test instances [8, 9].")
arg_parser.add_argument("--max_nr_packages", type=int, required=True, help="Maximum number of packages in an instance")
arg_parser.add_argument("--nr_instances_per_setup", type=int, required=True, help="Number of instances to generate for each setup")
args = arg_parser.parse_args()

grid_size_splits = args.grid_size_splits.split(",")
grid_size_splits = [int(size.strip()) for size in grid_size_splits]
# The smallest grid size is 2
grid_size_splits.insert(0, 1)
grid_size_intervals = [(grid_size_splits[i] + 1, grid_size_splits[i + 1]) for i in range(len(grid_size_splits) - 1)]
max_nr_packages = args.max_nr_packages
nr_instances_per_setup = args.nr_instances_per_setup

domain_file = "domain.pddl"
splits = ["train", "validation", "test"]

output_folder = "output"
if os.path.exists(output_folder):
    shutil.rmtree(output_folder)
os.makedirs(output_folder)

def _generate_instances(split: str, grid_size_interval: tuple[int, int], max_nr_packages: int, nr_instances_per_setup: int, domain_file: Path = None) -> None:
    split_folder = os.path.join(output_folder, split)
    os.makedirs(split_folder)

    if domain_file:
        # Copy the domain file to the split folder
        shutil.copy(domain_file, split_folder)
        print(f"Copied domain file to {split_folder}")

    for size in range(grid_size_interval[0], grid_size_interval[1] + 1):
        for nr_packages in range(1, max_nr_packages + 1):
            for index in range(nr_instances_per_setup):
                pddl_reader = PDDLReader(raise_on_error=True)
                pddl_reader.parse_domain(domain_file)
                problem = pddl_reader.problem
                language = problem.language
                init: Model = tarski.model.create(language)

                # Create the cells
                # All instances are square grids
                cells: list[Constant] = []
                for x in range(size):
                    for y in range(size):
                        cell = language.constant(f"c_{x}_{y}", "cell")
                        cells.append(cell)

                # Create the packages
                packages: list[Constant] = []
                for i in range(1, nr_packages + 1):
                    package = language.constant(f"p{i}", "package")
                    packages.append(package)

                # Create the truck
                truck = language.constant("t1", "truck")

                # Add the "adjacent" atoms
                adjacent_predicate = language.get_predicate("adjacent")
                for x in range(size):
                    for y in range(size):
                        current_cell = language.get_constant(f"c_{x}_{y}")
                        if x < size - 1:
                            right_neighbor = language.get_constant(f"c_{x + 1}_{y}")
                            init.add(adjacent_predicate, current_cell, right_neighbor)
                        if x > 0:
                            left_neighbor = language.get_constant(f"c_{x - 1}_{y}")
                            init.add(adjacent_predicate, current_cell, left_neighbor)
                        if y < size - 1:
                            bottom_neighbor = language.get_constant(f"c_{x}_{y + 1}")
                            init.add(adjacent_predicate, current_cell, bottom_neighbor)
                        if y > 0:
                            top_neighbor = language.get_constant(f"c_{x}_{y - 1}")
                            init.add(adjacent_predicate, current_cell, top_neighbor)

                # Randomly sample cells for the packages
                at_predicate = language.get_predicate("at")
                package_cell_names = []
                for package in packages:
                    cell = random.choice(cells)
                    package_cell_names.append(cell.name)
                    init.add(at_predicate, package, cell)

                # Randomly sample a cell for the truck
                cell = random.choice(cells)
                init.add(at_predicate, truck, cell)

                # The truck always starts out empty
                init.add(language.get_predicate("empty"), truck)

                # Create the goal
                # If all packages are in the same cell, make sure that this is not the goal cell
                if len(set(package_cell_names)) == 1:
                    goal_cell = random.choice(cells)
                    while goal_cell.name == package_cell_names[0]:
                        goal_cell = random.choice(cells)
                else:
                    goal_cell = random.choice(cells)
                goal_atoms = [at_predicate(package, goal_cell) for package in packages]

                problem.name = f"delivery-{size}x{size}-{nr_packages}"
                problem.init = init
                problem.goal = land(*goal_atoms)

                # Write the problem to a file
                writer = FstripsWriter(problem)
                output_file_name = f"instance_s-{size}_p-{nr_packages}_v-{index}.pddl"
                output_file_path = os.path.join(split_folder, output_file_name)
                writer.write_instance(output_file_path, constant_objects=[])
                print(f"Generated {output_file_path}")

for split, grid_size_interval in zip(splits, grid_size_intervals):
    _generate_instances(split, grid_size_interval, max_nr_packages, nr_instances_per_setup, domain_file=domain_file)

print(f"Generated Delivery instances in {output_folder}")



import json
import os
import sys
from collections import defaultdict
from functools import reduce
from itertools import cycle
from typing import List

import seaborn

# Add current project to sys path
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, parent_dir)
import logging
import pandas as pd
import matplotlib.pyplot as plt
plt.style.use('ggplot')


def plot_metric(metric: List[str],
                img_file_path: str) -> None:

    # get metrics for each algorithm and run
    algs = [a for a in os.listdir(RES_DIR) if '.' not in a]

    alg_stats = defaultdict(list)
    for alg in algs:
        for run in [d for d in os.listdir(os.path.join(RES_DIR, alg)) if '.DS_' not in d]:
            with open(f"{RES_DIR}/{alg}/{run}/metrics.json", 'r') as f:
                # alg_stats[alg].append(pd.json_normalize(json.load(f)))

                df = pd.DataFrame.from_dict(
                    {
                        domain: {
                            metric[-1]: reduce(lambda d, key: d.get(key) if d else None, metric, values),
                            # 'applicability': values.get('applicability').get('avg_recall'),
                        }
                        for domain, values in json.load(f).items()
                    },
                    orient='index'
                ).reset_index().rename(columns={'index': 'domain'})
                alg_stats[alg].append(df)

    # plot each algorithm metric
    domain_df = pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")
    avg_alg_stats = defaultdict()
    for alg in alg_stats:
        # extract numeric columns
        numeric_cols = alg_stats[alg][0].select_dtypes(include='number').columns

        # average only numeric parts
        avg_numeric_df = pd.concat([df[numeric_cols] for df in alg_stats[alg]]).groupby(level=0).mean()

        # merge non-numeric (i.e. 'domain') from the original dataframe
        non_numeric_cols = alg_stats[alg][0].drop(columns=numeric_cols)
        avg_metric = pd.concat([non_numeric_cols, avg_numeric_df], axis=1)

        # sort by domain dataframe column `operators`
        merged_df = avg_metric.merge(domain_df[['domain', 'operators']], on='domain')
        merged_df = merged_df.sort_values(by='operators')

        metric_series = merged_df.set_index('domain')[metric[-1]].astype(float)
        metric_series.name = alg  # gives the column a name

        avg_alg_stats[alg] = metric_series

    lineplot_df = pd.concat(avg_alg_stats.values(), axis=1)
    lineplot_df.columns = avg_alg_stats.keys()

    # Sort domains by average value across all algorithms
    domain_means = lineplot_df.mean(axis=1)
    lineplot_df = lineplot_df.loc[domain_means.sort_values().index]

    fig, ax = plt.subplots(figsize=(12, 6))

    for k, alg in enumerate(lineplot_df.columns):
        ax.plot(
            lineplot_df.index,
            lineplot_df[alg],
            label=alg,
            marker=next(markers),
            color=color_map[next(colors)],
            # alpha=.8,
            # linewidth=2,
            markersize=5 + (len(lineplot_df.columns) - k)*1.5
        )

    plt.legend(prop={'size': 13}, loc='lower right')

    # plt.title('Average number of objects')
    plt.xlabel('')
    plt.ylabel(f'{metric[-1].replace("_", " ").capitalize()}', size=18)
    plt.ylabel('')
    plt.title(' '.join(metric).replace('_', ' ').capitalize(), size=18)
    plt.xticks(rotation=70, size=13)
    plt.yticks(rotation=0, size=13)
    plt.tight_layout()
    plt.savefig(img_file_path)


def save_results_dataframe(res_file_path: str) -> None:
    # get metrics for each algorithm and run
    algs = [a for a in os.listdir(RES_DIR) if '.' not in a]

    alg_stats = defaultdict(list)
    for alg in algs:
        for run in [d for d in os.listdir(os.path.join(RES_DIR, alg)) if '.DS_' not in d]:
            alg_stats[alg].append(pd.read_excel(f"{RES_DIR}/{alg}/{run}/metrics.xlsx"))

    # Add a column to each
    all_dfs = []
    for alg, stats in alg_stats.items():
        alg_stats[alg][0]['algorithm'] = alg  # TODO: automatically average multiple runs
        all_dfs.append(alg_stats[alg][0])

    # Concatenate them
    df_all = pd.concat(all_dfs, ignore_index=True)

    # Sort by domain
    df_all = df_all.sort_values(by='domain')

    # Move "algorithm" column at the beginning
    df_all = df_all[[df_all.columns[-1]] + list(df_all.columns[:-1])]

    # Apply color gradient
    df_all = df_all.style.background_gradient(axis=0)

    # Save the dataframe to excel
    assert res_file_path.endswith('.xlsx'), f".xlsx file format must be used, current file name is {res_file_path}"
    df_all.to_excel(f"{RES_DIR}/{res_file_path}", index=False)


def lineplot_metric(metric: str,
                    img_file_path: str) -> None:
    # get metrics for each algorithm and run
    algs = [a for a in os.listdir(RES_DIR) if '.' not in a]

    alg_stats = defaultdict(list)
    for alg in algs:
        for run in [d for d in os.listdir(os.path.join(RES_DIR, alg)) if '.DS_' not in d]:
            alg_stats[alg].append(pd.read_excel(f"{RES_DIR}/{alg}/{run}/metrics.xlsx"))

    # plot each algorithm metric
    domain_df = pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")
    avg_alg_stats = defaultdict()
    for alg in alg_stats:
        # extract numeric columns
        numeric_cols = alg_stats[alg][0].select_dtypes(include='number').columns

        # average only numeric parts
        avg_numeric_df = pd.concat([df[numeric_cols] for df in alg_stats[alg]]).groupby(level=0).mean()

        # merge non-numeric (i.e. 'domain') from the original dataframe
        non_numeric_cols = alg_stats[alg][0].drop(columns=numeric_cols)
        avg_metric = pd.concat([non_numeric_cols, avg_numeric_df], axis=1)

        # sort by domain dataframe column `operators`
        merged_df = avg_metric.merge(domain_df[['domain', 'operators']], on='domain')
        merged_df = merged_df.sort_values(by='operators')

        metric_series = merged_df.set_index('domain')[metric]
        metric_series.name = alg  # gives the column a name

        avg_alg_stats[alg] = metric_series

    lineplot_df = pd.concat(avg_alg_stats.values(), axis=1)
    lineplot_df.columns = avg_alg_stats.keys()

    # Sort domains by average value across all algorithms
    domain_means = lineplot_df.mean(axis=1)
    lineplot_df = lineplot_df.loc[domain_means.sort_values().index]

    fig, ax = plt.subplots(figsize=(12, 6))

    for k, alg in enumerate(lineplot_df.columns):
        ax.plot(
            lineplot_df.index,
            lineplot_df[alg],
            label=alg,
            marker=next(markers),
            color=color_map[next(colors)],
            # alpha=.8,
            linewidth=2,
            markersize=5 + (len(lineplot_df.columns) - k)*1.5
        )

    plt.legend(prop={'size': 15}, loc='lower right')

    # plt.title('Average number of objects')
    plt.xlabel('')
    plt.ylabel('')
    plt.xticks(rotation=70, size=15)
    plt.yticks(rotation=0, size=15)
    plt.yticks([i * 0.1 for i in range(2, 11)],
               rotation=0, size=15)
    plt.tight_layout()
    plt.savefig(img_file_path)


def barplot_metric_traj_avg(metric: str,
                   img_file_path: str) -> None:

    # get metrics for each algorithm and run
    algs = [a for a in os.listdir(RES_DIR) if '.' not in a]
    # algs = ['ROSAME']

    # alg_stats = defaultdict(list)
    alg_stats = dict()
    for alg in algs:
        alg_stats[alg] = defaultdict(list)
        for n_traj in [d for d in os.listdir(os.path.join(RES_DIR, alg)) if '.DS_' not in d]:
            for run in [d for d in os.listdir(os.path.join(RES_DIR, alg, n_traj)) if '.DS_' not in d]:
                alg_stats[alg][n_traj].append(pd.read_excel(f"{RES_DIR}/{alg}/{n_traj}/{run}/metrics.xlsx"))

    # plot each algorithm metric
    domain_df = pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")
    # avoided = ['npuzzle', 'spanner', 'ferry', 'transport', 'miconic', 'sokoban', 'blocksworld']
    # avoided = ['miconic', 'blocksworld', 'satellite']
    # avoided = ['npuzzle', 'sokoban', 'transport', 'ferry', 'spanner', 'miconic', 'blocksworld', 'satellite']
    avoided = []
    domain_df = domain_df[~domain_df['domain'].isin(avoided)]
    avg_alg_stats = defaultdict()
    for alg in alg_stats:
        # extract numeric columns
        dummy_key = list(alg_stats[alg].keys())[0]
        numeric_cols = alg_stats[alg][dummy_key][0].select_dtypes(include='number').columns

        for n_traj in alg_stats[alg]:
            avg_alg_stats[n_traj] = defaultdict()
            # average only numeric parts
            avg_numeric_df = pd.concat([df[numeric_cols] for df in alg_stats[alg]]).groupby(level=0).mean()

            # merge non-numeric (i.e. 'domain') from the original dataframe
            non_numeric_cols = alg_stats[alg][0].drop(columns=numeric_cols)
            avg_metric = pd.concat([non_numeric_cols, avg_numeric_df], axis=1)

            # sort by domain dataframe column `operators`
            merged_df = avg_metric.merge(domain_df[['domain', 'operators']], on='domain')
            merged_df = merged_df.sort_values(by='operators')

            metric_series = merged_df.set_index('domain')[metric]
            metric_series.name = alg  # gives the column a name


            avg_alg_stats[n_traj][alg] = metric_series

    avg_metric_per_ntraj = avg_alg_stats[n_traj].values()
    barplot_df = pd.concat(avg_alg_stats.values(), axis=1)
    barplot_df.columns = avg_alg_stats.keys()

    # # # Sort domains by average value across all algorithms
    # domain_means = barplot_df.mean(axis=1)
    # barplot_df = barplot_df.loc[domain_means.sort_values().index]

    # plot grouped bar chart
    barplot_df.plot(
        # kind='barh',
        kind='bar',
        figsize=(12, 6),
        alpha=.8,
        color=[color_map[next(colors)] for _ in range(len(avg_alg_stats))]
    )

    plt.legend(prop={'size': 25}, loc='lower right')

    plt.xlabel('')
    plt.ylabel('Problem solving', size=25)
    # plt.ylabel('Pred. effs. Precision', size=25)
    plt.xticks(rotation=-30, ha='left', rotation_mode='anchor', size=25)
    plt.yticks(rotation=0, size=25)
    plt.ylim(.0, 1.)
    # plt.xticks([])
    plt.tight_layout()
    plt.savefig(img_file_path)

def barplot_metric(metric: str,
                   img_file_path: str) -> None:

    # get metrics for each algorithm and run
    algs = [a for a in os.listdir(RES_DIR) if '.' not in a]
    # algs = ['ROSAME']

    alg_stats = defaultdict(list)
    for alg in algs:
        for run in [d for d in os.listdir(os.path.join(RES_DIR, alg)) if '.DS_' not in d]:
            alg_stats[alg].append(pd.read_excel(f"{RES_DIR}/{alg}/{run}/metrics.xlsx"))

    # plot each algorithm metric
    domain_df = pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")
    # avoided = ['npuzzle', 'sokoban', 'transport', 'ferry', 'spanner', 'miconic', 'blocksworld', 'satellite']
    avoided = []
    domain_df = domain_df[~domain_df['domain'].isin(avoided)]
    avg_alg_stats = defaultdict()
    for alg in alg_stats:
        # extract numeric columns
        numeric_cols = alg_stats[alg][0].select_dtypes(include='number').columns

        # average only numeric parts
        avg_numeric_df = pd.concat([df[numeric_cols] for df in alg_stats[alg]]).groupby(level=0).mean()

        # merge non-numeric (i.e. 'domain') from the original dataframe
        non_numeric_cols = alg_stats[alg][0].drop(columns=numeric_cols)
        avg_metric = pd.concat([non_numeric_cols, avg_numeric_df], axis=1)

        # sort by domain dataframe column `operators`
        merged_df = avg_metric.merge(domain_df[['domain', 'operators']], on='domain')
        merged_df = merged_df.sort_values(by='operators')

        metric_series = merged_df.set_index('domain')[metric]
        metric_series.name = alg  # gives the column a name

        avg_alg_stats[alg] = metric_series

    barplot_df = pd.concat(avg_alg_stats.values(), axis=1)
    barplot_df.columns = avg_alg_stats.keys()

    # # # Sort domains by average value across all algorithms
    # domain_means = barplot_df.mean(axis=1)
    # barplot_df = barplot_df.loc[domain_means.sort_values().index]

    # plot grouped bar chart
    barplot_df.plot(
        # kind='barh',
        kind='bar',
        figsize=(12, 6),
        alpha=.8,
        color=[color_map[next(colors)] for _ in range(len(avg_alg_stats))]
    )

    plt.legend(prop={'size': 25}, loc='lower right')

    plt.xlabel('')
    plt.ylabel('Problem solving', size=25)
    # plt.ylabel('Pred. effs. Precision', size=25)
    plt.xticks(rotation=-30, ha='left', rotation_mode='anchor', size=25)
    plt.yticks(rotation=0, size=25)
    plt.ylim(.0, 1.)
    # plt.xticks([])
    plt.tight_layout()
    plt.savefig(img_file_path)


def print_metrics_table(metrics: List[str], run_dir: str = 'run0') -> None:
    # List algorithms
    algs = list(filter(os.path.isdir, [f"{RES_DIR}/{d}" for d in os.listdir(RES_DIR)]))

    # MultiIndex for columns
    columns = [('', 'Domain')] + [(metric, alg.split('/')[-1]) for metric in metrics for alg in algs]
    df = pd.DataFrame(columns=pd.MultiIndex.from_tuples(columns))

    # Load domains from the first algorithm
    with open(f'{algs[0]}/{run_dir}/metrics.json', 'r') as f:
        domains = json.load(f).keys()

    for d in domains:
        eval_row = {('', 'Domain'): d}
        metric_values = {metric: {} for metric in metrics}

        # Collect all values for later comparison
        for alg in algs:
            alg_name = alg.split('/')[-1]
            all_run_dfs = []
            for run_dir in os.listdir(f"{RES_DIR}/{alg}"):
                run_df = pd.read_excel(f"{RES_DIR}/{alg}/{run_dir}/metrics.xlsx")
                all_run_dfs.append(run_df)

            assert len(all_run_dfs) > 0, f'There is no run directory (e.g. "run0") in {RES_DIR}/{alg}.'

            # extract numeric columns
            numeric_cols = run_df.select_dtypes(include='number').columns

            # average only numeric parts
            avg_numeric_df = pd.concat([df[numeric_cols] for df in all_run_dfs]).groupby(level=0).mean()

            # merge non-numeric (i.e. 'domain') from the original dataframe
            non_numeric_cols = run_df.drop(columns=numeric_cols)
            alg_df = pd.concat([non_numeric_cols, avg_numeric_df], axis=1)

            alg_dom_df = alg_df[alg_df["domain"] == d]

            for metric in metrics:
                val = alg_dom_df[metric].values[0]
                metric_values[metric][alg_name] = val

        # Format with bold for best values
        for metric in metrics:
            values = metric_values[metric]
            max_val = max(values.values())
            min_val = min(values.values())

            for alg_name, val in values.items():
                val_fmt = f"{val:.2f}"
                if metric in ['false_plans_ratio']:
                    if val == min_val:
                        eval_row[(metric, alg_name)] = f"$\\mathbf{{{val_fmt}}}$"
                    else:
                        eval_row[(metric, alg_name)] = f"${val_fmt}$"
                else:
                    if val == max_val:
                        eval_row[(metric, alg_name)] = f"$\\mathbf{{{val_fmt}}}$"
                    else:
                        eval_row[(metric, alg_name)] = f"${val_fmt}$"

        df = pd.concat([df, pd.DataFrame([eval_row])], ignore_index=True)

    # # Transpose it
    # df = df.transpose()
    # # Optionally reset the index to make it a column in LaTeX output
    # df.reset_index(inplace=True)

    # Save to LaTeX
    df.to_latex(
        # f"{RES_DIR}/{'_'.join(metrics)}.tex",
        f"{RES_DIR}/all_metrics.tex",
        index=False,
        escape=False,
        # column_format=f"l|{'|'.join([''.join('c' for _ in algs) for _ in metrics])}",
        # label=f"tab:something",
        # caption="something"
    )


def print_best_table():

    # evaluated algorithms to be included in the table
    algs = ['SAM', 'OffLAM', 'ROSAME', 'NOLAM']

    # evaluation metrics to be included in the table
    metrics = ['domain', 'app precision', 'app recall',
               'predicted_effects precision', 'predicted_effects recall',
               'solving_ratio', 'false_plans_ratio']

    # get metrics for each algorithm averaged over all runs
    labeled_dfs = dict()
    for i, alg in enumerate(algs):
        all_run_dfs = []
        for run_dir in os.listdir(f"{RES_DIR}/{alg}"):
            run_df = pd.read_excel(f"{RES_DIR}/{alg}/{run_dir}/metrics.xlsx")
            all_run_dfs.append(run_df)

        assert len(all_run_dfs) > 0, f'There is no run directory (e.g. "run0") in {RES_DIR}/{alg}.'

        # extract numeric columns
        numeric_cols = run_df.select_dtypes(include='number').columns

        # average only numeric parts
        avg_numeric_df = pd.concat([df[numeric_cols] for df in all_run_dfs]).groupby(level=0).mean()

        # merge non-numeric (i.e. 'domain') from the original dataframe
        non_numeric_cols = run_df.drop(columns=numeric_cols)
        alg_df = pd.concat([non_numeric_cols, avg_numeric_df], axis=1)
        labeled_dfs[str(i + 1)] = alg_df[metrics]

    # sam = pd.read_excel(f"{RES_DIR}/SAM/{run}/metrics.xlsx")[metrics]
    # offlam = pd.read_excel(f"{RES_DIR}/OffLAM/{run}/metrics.xlsx")[metrics]
    # nolam = pd.read_excel(f"{RES_DIR}/NOLAM/{run}/metrics.xlsx")[metrics]
    # rosame = pd.read_excel(f"{RES_DIR}/ROSAME/{run}/metrics.xlsx")[metrics]
    #
    # # Label the dataframes
    # labeled_dfs = {'1': sam, '2': offlam, '3': rosame, '4': nolam}

    # Stack dataframes into a single multi-indexed dataframe
    combined = pd.concat(labeled_dfs, names=['Algorithm'])

    # Unstack so each column becomes a 3D array: (index, column, algorithm)
    stacked = combined.stack().unstack('Algorithm')

    # Compute the best value (e.g., max) for each cell
    best_values = stacked.max(axis=1)  # or min(axis=1) for minimization

    # Find labels that achieved the best value
    best_labels = stacked.eq(best_values, axis=0)

    # For each cell, get the list of algorithm names that matched the best value
    # labels_per_cell = best_labels.apply(lambda row: f"$^{{{','.join(row.index[row].tolist())}}}$", axis=1)
    labels_per_cell = best_labels.apply(lambda row: ','.join(row.index[row].tolist()), axis=1)

    # Combine best value and labels in a readable format
    result = best_values.astype(str) + ' $^{' + labels_per_cell + '}$'

    # Reshape back to original dataframe shape
    final_result = result.unstack()
    final_result.to_latex(f'{RES_DIR}/res_bests.tex', index=False, float_format="%.2f",
                          caption="Best metric values for every domain obtained among "
                                  f"{','.join([f'{i+1}: {k}' for i, k in enumerate(algs)])}"
                          )


if __name__ == '__main__':

    BENCHMARK_DIR = "benchmarks"

    RES_DIR = "../../res"

    logging.basicConfig(level=logging.INFO)

    colors = cycle(['purple', 'lightblue', 'gray', 'red', 'purple'])
    colors = cycle(['red', 'lightblue', 'purple', 'gray'])
    markers = cycle(['s', 'd', '*', '.'])

    # palette = seaborn.color_palette("Paired")
    # color_map = {
    #     "lightblue": palette[0],  # Default "lightblue" color in the pastel palette
    #     "blue": palette[1],  # Default "blue" color in the pastel palette
    #     "orange": palette[7],  # Default "orange" color
    #     "green": palette[3],   # ...
    #     "purple": palette[9],
    #     "brown": palette[11],
    #     "red": palette[5],
    # }
    palette = seaborn.color_palette("muted", 10)
    color_map = {
        "blue": palette[0],  # muted blue
        "orange": palette[1],  # muted orange
        "green": palette[2],  # muted green
        "red": palette[3],  # muted red
        "purple": palette[4],  # muted purple
        "brown": palette[5],  # muted brown
        "pink": palette[6],  # muted pink
        "gray": seaborn.color_palette("bright", 10)[-3],  # muted gray
        "yellow": palette[8],  # pale cyan-like
        "lightblue": palette[9],  # muted yellow
    }

    # Concatenate and save all algorithm results
    # save_results_dataframe('results.xlsx')

    # Generate metric barplot grouped by domain for every algorithm
    # barplot_metric('syn precision', f"{RES_DIR}/syn_precision.png")
    # barplot_metric('syn recall', f"{RES_DIR}/syn_recall.png")
    # barplot_metric('app precision', f"{RES_DIR}/app_precision.png")
    # barplot_metric('app recall', f"{RES_DIR}/app_recall.png")
    # barplot_metric('predicted_effects precision', f"{RES_DIR}/predeffs_precision.png")
    # barplot_metric('predicted_effects recall', f"{RES_DIR}/predeffs_recall.png")
    barplot_metric('solving_ratio', f"{RES_DIR}/solving.png")
    # barplot_metric('false_plans_ratio', f"{RES_DIR}/false_plans.png")

    # Generate metric lineplot grouped by domain for every algorithm
    # lineplot_metric('syn precision', f"{RES_DIR}/syn_precision_line.png")
    # lineplot_metric('syn recall', f"{RES_DIR}/syn_recall_line.png")
    # lineplot_metric('app precision', f"{RES_DIR}/app_precision_line.png")
    # lineplot_metric('app recall', f"{RES_DIR}/app_recall_line.png")
    # lineplot_metric('predicted_effects precision', f"{RES_DIR}/predeffs_precision_line.png")
    # lineplot_metric('predicted_effects recall', f"{RES_DIR}/predeffs_recall_line.png")
    # lineplot_metric('solving_ratio', f"{RES_DIR}/solving_line.png")
    # lineplot_metric('false_plans_ratio', f"{RES_DIR}/false_plans_line.png")

    # Print a table for the given metrics
    # metrics = [
    #     # "syn precision",
    #     # "syn recall",
    #     # "app precision",
    #     # "app recall",
    #     # "predicted_effects precision",
    #     # "predicted_effects recall",
    #     "solving_ratio",
    #     # "false_plans_ratio"
    # ]
    # print_metrics_table(metrics)

    # Print the best results achieved for every domain among all algorithms
    # print_best_table()

    # Plot metric results for preconditions/effects
    # for metric in [
    #     # ['syntactic', 'precs_pos_precision'],
    #     # ['syntactic', 'precs_neg_precision'],
    #     # ['syntactic', 'pos_precision'],
    #     # ['syntactic', 'neg_precision'],
    #     ['syntactic', 'precs_pos_recall'],
    #     # ['syntactic', 'pos_recall'],
    #     # ['syntactic', 'neg_recall']
    # ]:
    #     plot_metric(metric, f"{RES_DIR}/{metric[-1]}.png")



import abc
from typing import List, TypeVar

ActionType = TypeVar('ActionType')
StateType = TypeVar('StateType')


class Env(abc.ABC):
    """
    An AMLGym environment interface that is required for measuring predictive power
     metrics. An example of AMLGym environment is a unified-planning sequential simulator
     that implements the :meth:`apply` and :meth:`applicable_actions` methods.
    """

    @abc.abstractmethod
    def apply(self, s: StateType, a: ActionType) -> StateType:  # TODO: support stochastic transitions?
        """
        Return the state obtained after exeucuting action `a` in the current state `s`
        :param s: the current state `s`
        :param a: the action `a` to execute
        :return: the state `s'` reached after executing action `a` in state `s`
        """
        pass

    @abc.abstractmethod
    def applicable_actions(self, state: StateType) -> List[ActionType]:
        """
        Returns the set of actions applicable in the current state `s`
        :param state: the current state `s`
        :return: the list of actions applicable in the current state `s`
        """
        pass



import pddlgym
import imageio
import matplotlib.pyplot as plt
import numpy as np
import torch
import numpy.random as npr
import random
import argparse


objects = ['d1', 'd2', 'd3', 'd4', 'peg1', 'peg2', 'peg3']
action_order = [f'(move {obj1} {obj2} {obj3})' for obj1 in objects for obj2 in objects for obj3 in objects
                if obj1!=obj2 and obj1!=obj3 and obj2!=obj3]
action_order = {k: v for v, k in enumerate(action_order)}
propositions = [f'(clear {obj})' for obj in objects]
propositions.extend([f'(on {obj1} {obj2})' for obj1 in objects for obj2 in objects if obj1!=obj2])
discs = objects[:4]
propositions.extend([f'(smaller {obj1} {obj2})' for obj1 in discs for obj2 in discs if obj1!=obj2])
propositions = {k: v for v, k in enumerate(propositions)}

def obs_to_label(obs):
    label = np.zeros(len(propositions))
    for literal in obs.literals:
        if literal.pddl_str() not in propositions:
            # This is for handfull
            continue
        label[propositions[literal.pddl_str()]] = 1
    return label

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="-s save_addr")
    parser.add_argument("-s", default="data", help="save address")
    parser.add_argument("--seed", type=int, default=8800)
    args = parser.parse_args()

    torch.manual_seed(args.seed)
    random.seed(args.seed)
    npr.seed(args.seed)

    # Use PDDL Gym
    # https://github.com/tomsilver/pddlgym
    env = pddlgym.make("PDDLEnvHanoi_operator_actions-v0")
    env.fix_problem_index(1)
    obs, info = env.reset()

    labels = []
    actions = []

    # Create a long trace to be cut later
    for i in range(500):#5500
        imageio.imsave(f"{args.s}/{i}.png", env.render())
        plt.close()
        labels.append(obs_to_label(obs))
        action = env.action_space.sample(obs)
        actions.append(action_order[action.pddl_str()])
        obs = env.step(action)[0]

    labels_tensor = torch.tensor(np.array(labels), dtype=torch.float32)
    actions_tensor = torch.tensor(np.array(actions))

    torch.save(labels_tensor, f"{args.s}/labels.pt")
    torch.save(actions_tensor, f"{args.s}/actions.pt")




import os
import sys

import numpy as np
from alive_progress import alive_bar

# Add current project to sys path
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, parent_dir)
import logging
from unified_planning.io import PDDLReader
from collections import defaultdict
from typing import Dict, Tuple
import pandas as pd
import re
import matplotlib.pyplot as plt
plt.style.use('ggplot')


def get_object_stats(domain_file: str,
                     problem_file: str) -> Dict[Tuple[str, str], float]:

    # Parse the problem in unified-planning
    reader = PDDLReader()
    problem = reader.parse_problem(domain_file, problem_file)

    # Get number of objects grouped by type
    objects = defaultdict(int)
    objects['objects'] = len(problem.all_objects)  # get total number of objects
    for o in problem.all_objects:
        objects[o.type.name] += 1

    return objects


def write_trajectory_stats(file_name):
    all_domains_stats = dict()
    domains = sorted([d for d in os.listdir(f"../{BENCHMARK_DIR}/{TRAJ_DIR}")
                      if d not in ['.DS_Store']])

    min_objs = np.inf
    max_objs = -1
    min_states = np.inf
    max_states = -1
    with alive_bar(len(domains),
                   title=f'Producing trajectory statistics',
                   length=20,
                   bar='notes') as bar:

        for domain in domains:

            domain_file = f'../{BENCHMARK_DIR}/{DOMAINS_DIR}/{domain}.pddl'
            all_traj_stats = []

            for traj in sorted(os.listdir(f"../{BENCHMARK_DIR}/{TRAJ_DIR}/{domain}"),
                               key=lambda x: int(x.split("_")[0])):
                traj_stats = {'id': traj}

                traj_file = f"../{BENCHMARK_DIR}/{TRAJ_DIR}/{domain}/{traj}"
                problem_file = f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}/{traj.replace('_traj', '_prob.pddl')}"

                # Get number of states and actions
                with open(traj_file, 'r') as f:
                    rows = f.readlines()
                    traj_stats['actions'] = len([r for r in rows if r.startswith("(:action ")])
                    traj_stats['states'] = len([r for r in rows if r.startswith("(:state ")])

                # Get objects stats
                traj_stats.update(get_object_stats(domain_file, problem_file))

                # Store trajectory stats
                all_traj_stats.append(traj_stats)

            # Store domain stats
            stats = pd.DataFrame(all_traj_stats)
            all_domains_stats[domain] = stats

            # Update overall stats
            min_objs = min(min_objs, all_domains_stats[domain]['objects'].min())
            max_objs = max(max_objs, all_domains_stats[domain]['objects'].max())
            min_states = min(min_states, all_domains_stats[domain]['states'].min())
            max_states = max(max_states, all_domains_stats[domain]['states'].max())

            bar()  # update progress bar

    # Write all domains statistics into xlsx
    writer = pd.ExcelWriter(file_name, engine='xlsxwriter')

    # Define a header format
    header_format = writer.book.add_format(
        {
            "bold": True,
            "text_wrap": True,
            "valign": "top",
            "fg_color": "#D7E4BC",
            "border": 1,
        }
    )

    # Write one sheet for each domain
    for name, stats in all_domains_stats.items():
        stats.to_excel(writer, sheet_name=name, index=False, float_format="%.2f")

        # Auto-adjust the column width to make headers fully readable
        worksheet = writer.sheets[name]
        for col_num, value in enumerate(stats.columns.values):
            worksheet.write(0, col_num, value, header_format)
            max_len = max(
                stats[value].astype(str).map(len).max(),
                len(value)
            ) + 2  # offset
            worksheet.set_column(col_num, col_num, max_len)
    writer.close()

    # print overall stats
    print(f'Minimum number of objects: {min_objs}\n'
          f'Maximum number of objects: {max_objs}\n'
          f'Minimum number of states: {min_states}\n'
          f'Maximum number of states: {max_states}')


def write_domain_stats(file_name):

    all_domains_stats = list()
    domains = sorted([d for d in os.listdir(f"../{BENCHMARK_DIR}/{TRAJ_DIR}")
                          if d not in ['.DS_Store']])
    with alive_bar(len(domains),
                   title=f'Producing domain statistics',
                   length=20,
                   bar='bubbles') as bar:

        for domain in domains:

            # Parse the domain in unified-planning
            reader = PDDLReader()
            domain_file = f'../{BENCHMARK_DIR}/{DOMAINS_DIR}/{domain}.pddl'
            upl_domain = reader.parse_problem(domain_file)

            domain_stats = {
                'domain': domain,
                'operators': len(upl_domain.actions),
                'min operators arity': min([len(a.parameters) for a in upl_domain.actions]),
                'max operators arity': max([len(a.parameters) for a in upl_domain.actions]),
                'predicates': len(upl_domain.fluents),
                'min predicates arity': min(f.arity for f in upl_domain.fluents),
                'max predicates arity': max(f.arity for f in upl_domain.fluents),
                'types': len(upl_domain.user_types),
                'constants': 'no' if domain not in ['childsnack'] else 'yes',
                'negative preconditions': 'no',
                'injective assumption': 'no' if np.any([len([p.type.name for p in a.parameters])
                                                    > len({p.type.name for p in a.parameters})
                                                    for a in upl_domain.actions]) else 'yes'

            }

            # Store domain stats
            all_domains_stats.append(domain_stats)

            bar()  # update progress bar

    # Write all domains statistics into xlsx
    stats = pd.DataFrame(all_domains_stats)
    writer = pd.ExcelWriter(file_name, engine='xlsxwriter')

    # Define a header format
    header_format = writer.book.add_format(
        {
            "bold": True,
            "text_wrap": True,
            "valign": "top",
            "fg_color": "#D7E4BC",
            "border": 1,
        }
    )

    # Write one sheet for each domain
    stats.to_excel(writer, sheet_name='domains', index=False, float_format="%.2f")
    # Access the worksheet

    # Auto-adjust the column width to make headers fully readable
    worksheet = writer.sheets['domains']
    for col_num, value in enumerate(stats.columns.values):
        worksheet.write(0, col_num, value, header_format)
        max_len = max(
            stats[value].astype(str).map(len).max(),
            len(value)
        ) + 2  # offset
        worksheet.set_column(col_num, col_num, max_len)
    writer.close()


def plot_avg_attr(df_file_path: str,
                     attribute: str,
                     img_file_path: str) -> None:

    # Load all sheets into a dictionary of DataFrames
    all_sheets = pd.read_excel(df_file_path, sheet_name=None)

    # Collect all data in one list
    all_data = []

    for sheet_name, df in all_sheets.items():
        df = df.copy()

        # extract the trajectory number from 'id' (e.g. '0_blocksworld_traj' -> '0')
        df['x'] = df['id'].apply(lambda s: re.match(r'^(\d+)_', s).group(1) if re.match(r'^(\d+)_', s) else None)

        # Keep only necessary columns
        all_data.append(df[['x', attribute]])

    # Combine all domains into one DataFrame
    combined = pd.concat(all_data)

    # Group by 'x' and average total_objects
    grouped = combined.groupby('x')[attribute].mean().sort_index()

    # Plot
    plt.figure(figsize=(10, 6))
    grouped.plot(kind='bar', color='skyblue', edgecolor='none')
    # plt.title('Average number of objects')
    plt.xlabel('Trajectory', size=18)
    plt.ylabel(f'#{attribute.capitalize()}', size=18)
    plt.xticks(rotation=0, size=13)
    plt.yticks(rotation=0, size=13)
    plt.tight_layout()
    plt.savefig(img_file_path)


def print_domain_stats(domain_df):
    df = pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")
    print(f"Max operators: {df['operators'].max()}")
    print(f"min operators: {df['operators'].min()}")
    print(f"Max predicates: {df['predicates'].max()}")
    print(f"min predicates: {df['predicates'].min()}")
    print(f"Max types: {df['types'].max()}")
    print(f"min types: {df['types'].min()}")


def print_traj_table(traj_df):
    # Load Excel file with multiple sheets
    xls = pd.read_excel(traj_df, sheet_name=None)  # sheet_name=None loads all sheets

    # Identify common columns
    common_cols = set.intersection(*(set(df.columns) for df in xls.values()))

    # Compute averages for common columns per sheet
    averages = {
        sheet_name: df[list(common_cols)].mean(numeric_only=True)
        for sheet_name, df in xls.items()
    }

    # Convert to DataFrame for LaTeX export
    average_df = pd.DataFrame(averages).T  # Transpose to have sheet names as rows

    # Optionally round for presentation
    average_df = average_df.round(2)

    # Export to LaTeX
    latex_table = average_df.to_latex(index=True,
                                      caption="Stastistics of the trajectories.",
                                      label="tab:traj-stats",
                                      float_format="%.2f")

    print(latex_table)


if __name__ == '__main__':

    logging.basicConfig(level=logging.INFO)
    # logging.basicConfig(level=logging.DEBUG)
    BENCHMARK_DIR = "benchmarks"
    DOMAINS_DIR = "domains"
    PROB_DIR = "problems/learning"
    TRAJ_DIR = "trajectories/learning"

    # # Generate xlsx file of trajectory statistics
    # write_trajectory_stats(f"../{BENCHMARK_DIR}/trajectories.xlsx")
    # print_traj_table(f"../{BENCHMARK_DIR}/trajectories.xlsx")
    #
    # # Generate xlsx file of domain statistics
    # write_domain_stats(f"../{BENCHMARK_DIR}/domains.xlsx")
    #
    # # # Print domain statistics into a markdown table
    # domain_table_columns = ['domain', 'operators', 'predicates', 'types', 'constants',
    #                         'min operators arity', 'max operators arity',
    #                         'min predicates arity', 'max predicates arity', 'negative preconditions',
    #                         'injective assumption']
    # print(pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")[domain_table_columns].to_markdown(index=False))
    #
    # # Print domain statistics into an html table
    # print(pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")[domain_table_columns].to_html(index=False))
    #
    # # Print domain statistics into latex table
    # print(pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx")[domain_table_columns].to_latex(index=False))

    # # Print domain statistics
    # print_domain_stats(pd.read_excel(f"../{BENCHMARK_DIR}/domains.xlsx"))

    # # Plot number of objects for every trace averaged over all domains
    # plot_avg_attr(f"../{BENCHMARK_DIR}/trajectories_hard.xlsx",
    #               'objects',
    #               f"../{BENCHMARK_DIR}/objects.png")
    #
    # # Plot number of states for every trace averaged over all domains
    # plot_avg_attr(f"../{BENCHMARK_DIR}/trajectories_hard.xlsx",
    #               'states',
    #               f"../{BENCHMARK_DIR}/states.png")



#!/usr/bin/env python

import random
import itertools
import gen_random_tower_boards

# Generate all combinations from empty to each board.
# Usage ./generate-ipc-instances.py | bash

seed = 1
# Optimal
(size_x, size_y) = (4, 3)
for range_num_towers, range_height in [
    ([1, 2], range(3, 8)),
    ([3, 4], range(3, 6)),
    ([5, 6], range(3, 5)),
]:
    for height in range_height:
        for num_towers in range_num_towers:
            for i in range(4):
                scenario = gen_random_tower_boards.gen_board(
                    size_x, size_y, height, num_towers, seed
                )
                print(
                    "./generate.py boards/empty.txt boards/{} pddlfile --ensure_plan --store_plan --dont_remove_slack".format(
                        scenario
                    )
                )
                seed += 1
